#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "adaboost_glue.h"
#include <mlpack/methods/adaboost/adaboost.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::adaboost;

using namespace mlpack::tree;
using namespace mlpack::perceptron;
using namespace mlpack::util;

// Global Variable of the GlobalMethodObject object so it can be accessed from all functions
AdaBoost<> adaBoostPerceptron;
AdaBoost<DecisionTree<>> adaBoostDecisionStump;

bool usingPerceptron = true;
bool isPerceptronTrained = false;
bool isDecisionStumpTrained = false;

int trainedDataDimensionPerceptron = 0;
int trainedDataDimensionDecisionStump = 0;


// input: 	const MatType & 				data, 
//			const arma::Row< size_t > & 	labels,
//			const size_t 					numClasses,
//			const WeakLearnerType & 		other,
//			const size_t 					iterations = 100,
//			const double 					tolerance = 1e-6
// output: 
// description: 
//			Needs to be called first before all other predicates exept initModelNoTraining!
//			Initiates the Adaboostmodel and trains it, so classify can be used immediately.
//
void initModelWithTraining(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
							float *labelsArr, SP_integer labelsArrSize, 
							SP_integer numClasses, 
							char const *learner, 
							SP_integer iterations, double tolerance)
{
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if labels fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }

	// convert the Prolog array to arma::rowvec
	Row< size_t > labelsVector = convertArrayToVec(labelsArr, labelsArrSize);

	if(strcmp(learner, "perceptron") == 0)
	{
		usingPerceptron = true;
		adaBoostPerceptron = AdaBoost<Perceptron<>>(data, labelsVector, numClasses, perceptron::Perceptron<>(numClasses, 0UL, iterations), iterations, tolerance);
		isPerceptronTrained = true;
		trainedDataDimensionPerceptron = data.n_rows;
	}
	else if (strcmp(learner, "decision_stump") == 0)
	{
		usingPerceptron = false;
		adaBoostDecisionStump = AdaBoost<DecisionTree<>>(data, labelsVector, numClasses, tree::DecisionTree<>(data,labelsVector, numClasses), iterations, tolerance);
		isDecisionStumpTrained = true;
		trainedDataDimensionDecisionStump = data.n_rows;
	}
	else
	{
		raisePrologDomainExeption(learner, 2, "perceptron or decision_stump", "initModelWithTraining");
	}
}


// input: 	const double tolerance = 1e-6	
// output: 
// description: 
//			Needs to be called first before all other predicates exept initModelWithTraining!
//			Initiates the Adaboostmodel but doesn’t train it, so train has to be used first before classify can be used.
//
void initModelNoTraining(double tol = 1e-6, 
							char const *learner = "perceptron")
{
	if(strcmp(learner, "perceptron") == 0) 
	{
		usingPerceptron = true;
		adaBoostPerceptron = AdaBoost<Perceptron<>>(tol);
		isPerceptronTrained = false;
		trainedDataDimensionPerceptron = 0;
	}
	else if (strcmp(learner, "decision_stump") == 0)
	{
		usingPerceptron = false;
		adaBoostDecisionStump = AdaBoost<DecisionStump<>>(tol);
		isDecisionStumpTrained = false;
		trainedDataDimensionDecisionStump = 0;
	}
	else
	{
		raisePrologDomainExeption(learner, 2, "perceptron or decision_stump", "initModelNoTraining");
	}
}


// input: 	const MatType & 		test,
//			arma::Row< size_t > & 	predictedLabels <-,
//			arma::mat & 			probabilities <-
// output: 
// description: 
//			Classifies the given data into the number of classes the model was trained for.
//
void classify(float *testMatArr, SP_integer testMatSize, SP_integer testMatRowNum, 
				float **predicLabelsArr, SP_integer *predicLabelsArrSize, 
				float **probsMatArr, SP_integer *probsMatColNum, SP_integer *probsMatRowNum)
{
	// convert the Prolog array to arma::mat
	mat test = convertArrayToMat(testMatArr, testMatSize, testMatRowNum);

	// create the ReturnVector
	Row< size_t > predicLabelsReturnVector;
	// create the ReturnMat
	mat probsReturnMat;


	// run the classify function on the model depending on if perceptron was selected
	if(usingPerceptron && isPerceptronTrained)
	{
		if (trainedDataDimensionPerceptron != test.n_rows)
		{
			raisePrologSystemExeption("The given Datapoints Dimension is diffrent than the trained Dimension!");
			return;
		}
		try
		{
			adaBoostPerceptron.Classify(test, predicLabelsReturnVector, probsReturnMat);
		}
		catch(const std::exception& e)
		{
			raisePrologSystemExeption(e.what());
			return;
		}
		
		
	}
	else if(!usingPerceptron && isDecisionStumpTrained)
	{
		if (trainedDataDimensionDecisionStump != test.n_rows)
		{
			raisePrologSystemExeption("The given Datapoints Dimension is diffrent than the trained Dimension!");
			return;
		}
		try
		{
			adaBoostDecisionStump.Classify(test, predicLabelsReturnVector, probsReturnMat);
		}
		catch(const std::exception& e)
		{
			raisePrologSystemExeption(e.what());
			return;
		}
	}
	else
	{
		// raise an exeption when classify gets called before the model is trained
		raisePrologSystemExeption("The model is not trained!");
		return;
	}

	// check if the return vectors have actual numbers stored
	if (isnan(probsReturnMat[0]) || isnan(predicLabelsReturnVector[0]))
	{
		raisePrologSystemExeption("The given data matrix has incorrect dimensions compared to the training data!");
		return;
	}

	// return the Vector
	returnVectorInformation(predicLabelsReturnVector, predicLabelsArr, predicLabelsArrSize);
	/// return the Matrix
	returnMatrixInformation(probsReturnMat, probsMatArr, probsMatColNum, probsMatRowNum);
}


// input: 
// output: 	size_t 	classesNum
// description: 
//			Returns the amount of classes defined in the model for classification.
//
SP_integer numClasses()
{
	if(usingPerceptron) 
	{
		return adaBoostPerceptron.NumClasses();
	}
	else
	{
		return adaBoostDecisionStump.NumClasses();
	}
}


// input: 
// output: 	double 	tol
// description: 
//			Returns the tolerance of the model.
//
double getTolerance()
{
	if(usingPerceptron) 
	{
		return adaBoostPerceptron.Tolerance();
	}
	else
	{
		return adaBoostDecisionStump.Tolerance();
	}
}


// input: 
// output: 	double&	tol
// description: 
//			Modifies the tolerance of the model.
//
void modifyTolerance(double newTol)
{
	if(usingPerceptron) 
	{
		double& tol = adaBoostPerceptron.Tolerance();
		tol = newTol;
	}
	else
	{
		double& tol = adaBoostDecisionStump.Tolerance();
		tol = newTol;
	}
}


// input: 	const MatType & 				data,
//			const arma::Row< size_t > & 	labels,
//			const size_t 					numClasses,
//			const WeakLearnerType & 		learner,
//			const size_t 					iterations = 100,
//			const double 					tolerance = 1e-6
//
// output: 	double upper bound training error
// description: 
//			Trains the model with the given data.
//
double train(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
				float *labelsArr, SP_integer labelsArrSize, 
				SP_integer numClasses, 
				char const *learner, 
				SP_integer iterations, double tolerance)
{
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if labels fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }

	// convert the Prolog array to arma::rowvec
	Row< size_t > labelsVector = convertArrayToVec(labelsArr, labelsArrSize);


	// run the train fuction on the model given what learner was selected
	if(strcmp(learner, "perceptron") == 0 && usingPerceptron) 
	{
		try
		{
			isPerceptronTrained = true;
			double error = adaBoostPerceptron.Train(data, labelsVector, numClasses, Perceptron(), iterations, tolerance);
			trainedDataDimensionPerceptron = data.n_rows;
			return error;
		}
		catch(const std::exception& e)
		{
			if (strcmp(e.what(),"Mat::col(): index out of bounds") == 0)
			{
				raisePrologSystemExeption("The values of the Label have to start at 0 and be >= 0 and < the given numClass!");
			}
			else
			{
				raisePrologSystemExeption(e.what());
			}
			return 0;
		}
		
	}
	else if (strcmp(learner, "decision_stump") == 0 && !usingPerceptron)
	{
		try
		{
			isDecisionStumpTrained = true;
			double error = adaBoostDecisionStump.Train(data, labelsVector, numClasses, DecisionStump<>(), iterations, tolerance);
			trainedDataDimensionDecisionStump = data.n_rows;
			return error;
		}
		catch(const std::exception& e)
		{
			if (strcmp(e.what(),"Mat::col(): index out of bounds") == 0)
			{
				raisePrologSystemExeption("The values of the Label have to start at 0 and be >= 0 and < the given numClass!");
			}
			else
			{
				raisePrologSystemExeption(e.what());
			}
			return 0;
		}
	}
	else
	{
		raisePrologDomainExeption(learner, 5, "perceptron or decision_stump", "train");
	}
	return 0;
}
