
:- module(adaboost, [   adaboost_initModelWithTraining/7,
                        adaboost_initModelNoTraining/2, 
                        adaboost_classify/5, 
                        adaboost_numClasses/1, 
                        adaboost_getTolerance/1, 
                        adaboost_modifyTolerance/1, 
                        adaboost_train/8]).

%% requirements of library(struct)
:- load_files(library(str_decl),
                [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function


%% --Input--
%%            mat    data,
%%            vec    labels,
%%            int    numClasses,
%%            string learner       => "decision_stump", "perceptron",
%%            int    iterations    => 100,
%%            float  tolerance     => 1e-6
%%
%% --Output--
%%
%% --Description--
%%            Needs to be called first before all other predicates exept initModelNoTraining!
%%
%%            Initiates the Adaboostmodel and trains it, so classify can be used immediately.
%%
adaboost_initModelWithTraining(MatList, MatRows, VecList, NumClasses, Learner, Iterations, Tolerance) :-
        NumClasses >= 0,
        Iterations >= 0,
        Tolerance > 0,
        convert_list_to_float_array(MatList, MatRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(VecList, array(Ysize, Y)),
        initModelWithTrainingI(X, Xsize, Xrownum, Y, Ysize, NumClasses, Learner, Iterations, Tolerance).

foreign(initModelWithTraining, c, initModelWithTrainingI(       +pointer(float_array), +integer, +integer, 
                                                                +pointer(float_array), +integer, 
                                                                +integer, 
                                                                +string, 
                                                                +integer , +float32)).



%% --Input--
%%            float  tolerance     => 1e-6
%%            string learner       => "decision_stump", "perceptron",
%%
%% --Output--
%%
%% --Description--
%%            Needs to be called first before all other predicates exept initModelWithTraining!
%%            Initiates the Adaboostmodel but doesnt train it, so train has to be used first before classify can be used.
%%
adaboost_initModelNoTraining(Tolerance, Learner) :-
        Tolerance > 0,
        initModelNoTrainingI(Tolerance, Learner).

foreign(initModelNoTraining, c, initModelNoTrainingI(+float32, +string)).


%% --Input--
%%            mat    test,
%%
%% --Output--
%%            vec    predictions,
%%            mat    probabilities
%%
%% --Description--
%%            Classifies the given data into the number of classes the model was trained for.
%%
adaboost_classify(TestList, TestRows, PredicList, ProbsList, ZRows) :-
        convert_list_to_float_array(TestList, TestRows, array(Xsize, Xrownum, X)),
        classifyI(X, Xsize, Xrownum, Y, Ysize, Z, ZCols, ZRows),
        convert_float_array_to_list(Y, Ysize, PredicList),
        convert_float_array_to_2d_list(Z, ZCols, ZRows, ProbsList).

foreign(classify, c, classifyI( +pointer(float_array), +integer, +integer, 
                                -pointer(float_array), -integer, 
                                -pointer(float_array), -integer, -integer)).




%% --Input--
%%
%% --Output--
%%            float  classesNum
%%
%% --Description--
%%            Returns the amount of classes defined in the model for classification.
%%
adaboost_numClasses(ClassesNum) :-
        numClassesI(ClassesNum).

foreign(numClasses, c, numClassesI([-integer])).


%% --Input--
%%
%% --Output--
%%            float  tolerance
%%
%% --Description--
%%            Returns the tolerance of the model.
%%
adaboost_getTolerance(Tolerance) :-
        getToleranceI(Tolerance).

foreign(getTolerance, c, getToleranceI([-float32])).


%% --Input--
%%            float  newTolerance
%%
%% --Output--
%%
%% --Description--
%%            Modifies the tolerance of the model.
%%
adaboost_modifyTolerance(NewTolerance) :-
        NewTolerance > 0,
        modifyToleranceI(NewTolerance).

foreign(modifyTolerance, c, modifyToleranceI(+float32)).


%% --Input--
%%            mat    data,
%%            vec    labels,
%%            int    numClasses,
%%            string learner       => "decision_stump", "perceptron",
%%            int    iterations    => 100,
%%            float  tolerance     => 1e-6
%%
%% --Output--
%%            float  double upper bound training error
%%
%% --Description--
adaboost_train(MatList, MatRows, VecList, NumClasses, Learner, Iterations, Tolerance, Error) :-
        NumClasses >= 0,
        Iterations >= 0,
        Tolerance >= 0,
        convert_list_to_float_array(MatList, MatRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(VecList, array(Ysize, Y)),
        trainI(X, Xsize, Xrownum, Y, Ysize, NumClasses, Learner, Iterations, Tolerance, Error).

foreign(train, c, trainI(   +pointer(float_array), +integer, +integer, 
                            +pointer(float_array), +integer, 
                            +integer, 
                            +string, 
                            +integer , +float32, 
                            [-float32])).




%% Defines the functions that get connected from main.cpp
foreign_resource(adaboost, [    initModelWithTraining, 
                                initModelNoTraining, 
                                classify, 
                                numClasses, 
                                getTolerance, 
                                modifyTolerance, 
                                train]).

:- load_foreign_resource(adaboost).
