#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "approx_kfn_glue.h"
#include <mlpack/methods/approx_kfn/drusilla_select.hpp>
#include <mlpack/methods/approx_kfn/qdafn.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::neighbor;

// Global Variable of the DrusillaSelect object so it can be accessed from all functions
DrusillaSelect<mat> drusillaSelect = DrusillaSelect(1,1);
bool isDrusillaTrained = false;

// Global Variable of the QDAFN object so it can be accessed from all functions
QDAFN<mat> qdafn = QDAFN(1,1);
bool isQDAFNTrained = false;


// input:   const size_t 	l,
//          const size_t 	m
// output: 
// description: 
//          Initiates the DrusillaSearch Model but doesnt train it.
//
void initDrusillaModelNoTrain(SP_integer l, SP_integer m)
{
    drusillaSelect = DrusillaSelect(l, m);
    isDrusillaTrained = false;
}


// input:   const MatType & 	referenceSet,
//          const size_t 	    l,
//          const size_t 	    m
// output: 
// description: 
//          Initiates the DrusillaSearch Model and trains it with the given reference Set.
//
void initDrusillaModelWithTrain(float *referenceSetMatArr, SP_integer referenceSetMatSize, SP_integer referenceSetMatRowNum, 
                                SP_integer l, SP_integer m)
{
    // convert the Prolog arrays to arma::mat
    mat referenceSet = convertArrayToMat(referenceSetMatArr, referenceSetMatSize, referenceSetMatRowNum);
    
    
    try
    {
        drusillaSelect = DrusillaSelect(referenceSet, l, m);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    isDrusillaTrained = true;
}


// input:   const MatType & 	    querySet,
//          const size_t 	        k,
//          arma::Mat< size_t > & 	neighbors,
//          arma::mat & 	        distances
// output: 
// description: 
//          Run Search on the given Queryset with the Drusilla Search Policy.
//
void searchDrusilla(float *querySetMatArr, SP_integer querySetMatSize, SP_integer querySetMatRowNum, 
                    SP_integer k,
                    float **neighborsMatArr, SP_integer *neighborsMatColNum, SP_integer *neighborsMatRowNum, 
                    float **distancesMatArr, SP_integer *distancesMatColNum, SP_integer *distancesMatRowNum)
{
    if (!isDrusillaTrained)
    {
        raisePrologSystemExeption("The Model is not Trained!");
        return;
    }
    
    // convert the Prolog arrays to arma::mat
    mat querySet = convertArrayToMat(querySetMatArr, querySetMatSize, querySetMatRowNum);

    // create the ReturnMat
    Mat< size_t > neighborsReturnMat;
    mat distancesReturnMat;
    

    try
    {
        drusillaSelect.Search(querySet, k, neighborsReturnMat, distancesReturnMat);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    
    
    // return the Matrices
    returnMatrixInformation(neighborsReturnMat, neighborsMatArr, neighborsMatColNum, neighborsMatRowNum);
    returnMatrixInformation(distancesReturnMat, distancesMatArr, distancesMatColNum, distancesMatRowNum);
    
}


// input:   const MatType & 	referenceSet,
//          const size_t 	    l = 0,
//          const size_t 	    m = 0 
// output: 
// description: 
//          Trains the DrusillaSearch Model with the given reference Set.
//
void trainDrusilla(float *referenceSetMatArr, SP_integer referenceSetMatSize, SP_integer referenceSetMatRowNum, 
                    SP_integer l, SP_integer m)
{
    // convert the Prolog arrays to arma::mat
    mat referenceSet = convertArrayToMat(referenceSetMatArr, referenceSetMatSize, referenceSetMatRowNum);
    
    
    try
    {
        drusillaSelect.Train(referenceSet, l, m);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    isDrusillaTrained = true;
}


// input:   const size_t 	l,
//          const size_t 	m
// output: 
// description: 
//          Initiates the QDAFNSearch Model but doesnt train it.
//
void initQDAFNModelNoTrain(SP_integer l, SP_integer m)
{
    qdafn = QDAFN(l, m);
    isQDAFNTrained = false;
}


// input:   const MatType & 	referenceSet,
//          const size_t 	    l,
//          const size_t 	    m
// output: 
// description: 
//          Initiates the QDAFNSearch Model and trains it with the given reference Set.
//
void initQDAFNModelWithTrain(float *referenceSetMatArr, SP_integer referenceSetMatSize, SP_integer referenceSetMatRowNum, 
                                SP_integer l, SP_integer m)
{
    // convert the Prolog arrays to arma::mat
    mat referenceSet = convertArrayToMat(referenceSetMatArr, referenceSetMatSize, referenceSetMatRowNum);
    
    
    try
    {
        qdafn = QDAFN(referenceSet, l, m);
    }
    catch(const std::out_of_range& e)
    {
        raisePrologSystemExeption("l*m must be smaller than the number of points in the dataset.");
        return;
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    isQDAFNTrained = true;
}


// input:   const MatType & 	    querySet,
//          const size_t 	        k,
//          arma::Mat< size_t > & 	neighbors,
//          arma::mat & 	        distances
// output: 
// description: 
//          Run Search on the given Queryset with the QDAFN Search Policy.
//
void searchQDAFN(float *querySetMatArr, SP_integer querySetMatSize, SP_integer querySetMatRowNum, 
                    SP_integer k,
                    float **neighborsMatArr, SP_integer *neighborsMatColNum, SP_integer *neighborsMatRowNum, 
                    float **distancesMatArr, SP_integer *distancesMatColNum, SP_integer *distancesMatRowNum)
{
    if (!isQDAFNTrained)
    {
        raisePrologSystemExeption("The Model is not Trained!");
        return;
    }

    // convert the Prolog arrays to arma::mat
    mat querySet = convertArrayToMat(querySetMatArr, querySetMatSize, querySetMatRowNum);

    // create the ReturnMat
    Mat< size_t > neighborsReturnMat;
    mat distancesReturnMat;
    

    try
    {
        qdafn.Search(querySet, k, neighborsReturnMat, distancesReturnMat);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    
    
    // return the Matrices
    returnMatrixInformation(neighborsReturnMat, neighborsMatArr, neighborsMatColNum, neighborsMatRowNum);
    returnMatrixInformation(distancesReturnMat, distancesMatArr, distancesMatColNum, distancesMatRowNum);
}


// input:   const MatType & 	referenceSet,
//          const size_t 	    l = 0,
//          const size_t 	    m = 0 
// output: 
// description: 
//          Trains the QDAFNSearch Model with the given reference Set.
//
void trainQDAFN(float *referenceSetMatArr, SP_integer referenceSetMatSize, SP_integer referenceSetMatRowNum, 
                SP_integer l, SP_integer m)
{
    // convert the Prolog arrays to arma::mat
    mat referenceSet = convertArrayToMat(referenceSetMatArr, referenceSetMatSize, referenceSetMatRowNum);
    
    
    try
    {
        qdafn.Train(referenceSet, l, m);
    }
    catch(const std::out_of_range& e)
    {
        raisePrologSystemExeption("l*m must be smaller than the number of points in the dataset.");
        return;
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    isQDAFNTrained = true;
}
