
:- module(approx_kfn, [ initDrusillaModelNoTrain/2,
                        initDrusillaModelWithTrain/4,
                        searchDrusilla/7,
                        trainDrusilla/4,
                        initQDAFNModelNoTrain/2,
                        initQDAFNModelWithTrain/4,
                        searchQDAFN/7,
                        trainQDAFN/4]).

%% requirements of library(struct)
:- load_files(library(str_decl),
        [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function


%% --Input--
%%              int     l,
%%              int     m
%%
%% --Output--
%%
%% --Description--
%%              Initiates the DrusillaSearch Model but doesnt train it.
%%              trainDrusilla has to be used before searchDrusilla can be used.
%%
initDrusillaModelNoTrain(L, M) :-
        L > 0,
        M > 0,
        initDrusillaModelNoTrainI(L, M).

foreign(initDrusillaModelNoTrain, c, initDrusillaModelNoTrainI(+integer, +integer)).


%% --Input--
%%              mat     referenceSet,
%%              int     l,
%%              int     m
%%
%% --Output--
%%
%% --Description--
%%              Initiates the DrusillaSearch Model and trains it with the given reference Set.
%%              Afterwards searchDrusilla can be used.
%%
initDrusillaModelWithTrain(DataList, DataRows, L, M) :-
        L > 0,
        M > 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        initDrusillaModelWithTrainI(X, Xsize, Xrownum, L, M).
        
foreign(initDrusillaModelWithTrain, c, initDrusillaModelWithTrainI(     +pointer(float_array), +integer, +integer, 
                                                                        +integer, +integer)).


%% --Input--
%%              mat     querySet,
%%              int     k
%%
%% --Output--
%%              mat     neighbors,
%%              mat     distances
%%
%% --Description--
%%              Run Search on the given Queryset with the Drusilla Search Policy.
%%
searchDrusilla(DataList, DataRows, K, NeighborsList, YCols, DistancesList, ZCols) :-
        K > 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        searchDrusillaI(X, Xsize, Xrownum, K, Y, YCols, YRows, Z, ZCols, ZRows),
        convert_float_array_to_2d_list(Y, YCols, YRows, NeighborsList),
        convert_float_array_to_2d_list(Z, ZCols, ZRows, DistancesList).

foreign(searchDrusilla, c, searchDrusillaI(     +pointer(float_array), +integer, +integer, 
                                                +integer, 
                                                -pointer(float_array), -integer, -integer, 
                                                -pointer(float_array), -integer, -integer)).


%% --Input--
%%              mat     referenceSet,
%%              int     l               => 0,
%%              int     m               => 0
%%
%% --Output--
%%
%% --Description--
%%              Trains the DrusillaSearch Model with the given reference Set.
%%
trainDrusilla(DataList, DataRows, L, M) :-
        L > 0,
        M > 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        trainDrusillaI(X, Xsize, Xrownum, L, M).

foreign(trainDrusilla, c, trainDrusillaI(       +pointer(float_array), +integer, +integer, 
                                                +integer, +integer)).


%% --Input--
%%              int     l,
%%              int     m
%%
%% --Output--
%%
%% --Description--
%%              Initiates the QDAFNSearch Model but doesnt train it.
%%              trainQDAFN has to be used before searchQDAFN can be used.
%%
initQDAFNModelNoTrain(L, M) :-
        L > 0,
        M > 0,
        initQDAFNModelNoTrainI(L, M).

foreign(initQDAFNModelNoTrain, c, initQDAFNModelNoTrainI(+integer, +integer)).


%% --Input--
%%              mat     referenceSet,
%%              int     l,
%%              int     m
%%
%% --Output--
%%
%% --Description--
%%              Initiates the QDAFNSearch Model and trains it with the given reference Set.
%%              Afterwards searchQDAFN can be used.
%%
initQDAFNModelWithTrain(DataList, DataRows, L, M) :-
        L > 0,
        M > 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        initQDAFNModelWithTrainI(X, Xsize, Xrownum, L, M).

foreign(initQDAFNModelWithTrain, c, initQDAFNModelWithTrainI(   +pointer(float_array), +integer, +integer, 
                                                                +integer, +integer)).


%% --Input--
%%              mat     querySet,
%%              int     k
%%
%% --Output--
%%              mat     neighbors,
%%              mat     distances
%%
%% --Description--
%%              Run Search on the given Queryset with the QDAFN Search Policy.
%%              
searchQDAFN(DataList, DataRows, K, NeighborsList, YCols, DistancesList, ZCols) :-
        K > 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        searchQDAFNI(X, Xsize, Xrownum, K, Y, YCols, YRows, Z, ZCols, ZRows),
        convert_float_array_to_2d_list(Y, YCols, YRows, NeighborsList),
        convert_float_array_to_2d_list(Z, ZCols, ZRows, DistancesList).

foreign(searchQDAFN, c, searchQDAFNI(   +pointer(float_array), +integer, +integer, 
                                        +integer, 
                                        -pointer(float_array), -integer, -integer, 
                                        -pointer(float_array), -integer, -integer)).


%% --Input--
%%              mat     referenceSet,
%%              int     l               => 0,
%%              int     m               => 0
%%
%% --Output--
%%
%% --Description--
%%              Trains the QDAFNSearch Model with the given reference Set.
%%
trainQDAFN(DataList, DataRows, L, M) :-
        L > 0,
        M > 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        trainQDAFNI(X, Xsize, Xrownum, L, M).

foreign(trainQDAFN, c, trainQDAFNI(     +pointer(float_array), +integer, +integer, 
                                        +integer, +integer)).


%% Defines the functions that get connected from main.cpp
foreign_resource(approx_kfn, [  initDrusillaModelNoTrain,
                                initDrusillaModelWithTrain,
                                searchDrusilla,
                                trainDrusilla,
                                initQDAFNModelNoTrain,
                                initQDAFNModelWithTrain,
                                searchQDAFN,
                                trainQDAFN]).

:- load_foreign_resource(approx_kfn).
