#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "bayesian_linear_regression_glue.h"
#include <mlpack/methods/bayesian_linear_regression/bayesian_linear_regression.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::regression;

// Global Variable of the BayesianLinearRegression object so it can be accessed from all functions
BayesianLinearRegression regressor;
bool isModelTrained = false;

// input:	const bool 		centerData, 
//			const bool 		scaleData, 
//			const size_t 	nIterMax, 
//			const double 	tol
// output: 
void initModel(SP_integer centerData, SP_integer scaleData, SP_integer nIterMax, double tol)
{
	regressor = new BayesianLinearRegression((centerData == 1), (scaleData == 1), nIterMax, tol);
	isModelTrained = false;
}

// input:
// output: double
double alpha()
{
	return regressor.Alpha();
}

// input:
// output: double
double beta()
{
	return regressor.Beta();
}

// input:
// output: const arma::colvec &	dataOffset
void dataOffset(float **dataOffsetArr, SP_integer *dataOffsetArrSize)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not Trained!");
		return;
	}
	// create the ReturnVector
	vec dataOffsetReturnVector = regressor.DataOffset();
	
	// return the Vector
	returnVectorInformation(dataOffsetReturnVector, dataOffsetArr, dataOffsetArrSize);
}

// input:
// output: const arma::colvec &	dataScale
void dataScale(float **dataScaleArr, SP_integer *dataScaleArrSize)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not Trained!");
		return;
	}
	// create the ReturnVector
	vec dataScaleReturnVector = regressor.DataScale();
	
	// return the Vector
	returnVectorInformation(dataScaleReturnVector, dataScaleArr, dataScaleArrSize);
}

// input:
// output: const arma::colvec &	omega
void omega(float **omegaArr, SP_integer *omegaArrSize)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not Trained!");
		return;
	}
	// create the ReturnVector
	vec omegaReturnVector = regressor.Omega();
	
	// return the Vector
	returnVectorInformation(omegaReturnVector, omegaArr, omegaArrSize);
}

// input: 	const arma::mat &points, 
//			arma::rowvec &predictions
// output:
void predict(float *pointsMatArr, SP_integer pointsMatSize, SP_integer pointsMatRowNum, 
				float **predictionsArr, SP_integer *predictionsArrSize)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not Trained!");
		return;
	}
	// convert the Prolog array to arma::mat
	mat points = convertArrayToMat(pointsMatArr, pointsMatSize, pointsMatRowNum);
	
	// create the ReturnVector
	rowvec predictionsReturnVector;
	

	try
	{
		// run the prediction and save the result in arma::rowvec
		regressor.Predict(points, predictionsReturnVector);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}

	cout << points << endl;
	// return the Vector
	returnVectorInformation(predictionsReturnVector, predictionsArr, predictionsArrSize);
}

// input:	const arma::mat &points, 
//			arma::rowvec &predictions, 
//			arma::rowvec &std
// output:
void predictWithStd(float *pointsMatArr, SP_integer pointsMatSize, SP_integer pointsMatRowNum, 
					float **predictionsArr, SP_integer *predictionsArrSize, 
					float **stdArr, SP_integer *stdArrSize)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not Trained!");
		return;
	}
	// convert the Prolog array to arma::mat
	mat points = convertArrayToMat(pointsMatArr, pointsMatSize, pointsMatRowNum);
	
	// create the ReturnVectors
	rowvec predictionsReturnVector;
	rowvec stdReturnVector;
	

	try
	{
		// run the prediction and save the result in arma::rowvec
		regressor.Predict(points, predictionsReturnVector, stdReturnVector);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}


	// return the Vectors
	returnVectorInformation(predictionsReturnVector, predictionsArr, predictionsArrSize);
	returnVectorInformation(stdReturnVector, stdArr, stdArrSize);
}


// input:
// output: double
double responsesOffset()
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not Trained!");
		return 0.0;
	}
	return regressor.ResponsesOffset();
}

// input:	const arma::mat &		data, 
//			const arma::rowvec &	responses
// output: double
double rmse(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
			float *responsesArr, SP_integer responsesArrSize)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not Trained!");
		return 0.0;
	}
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	
	// convert the Prolog array to arma::rowvec
	rowvec responsesVector = convertArrayToRowvec(responsesArr, responsesArrSize);
	

	try
	{
		// run the model function and return the error
		return regressor.RMSE(data, responsesVector);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return 0.0;
	}
}

// input:	const arma::mat &		data, 
//			const arma::rowvec &	responses
// output:
void train(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
			float *responsesArr, SP_integer responsesArrSize)
{
	if(dataMatSize / dataMatRowNum != responsesArrSize)
	{
		raisePrologSystemExeption("Target dim doesnt fit to the Data dim");
		return;
	}
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	
	// convert the Prolog array to arma::rowvec
	rowvec responsesVector = convertArrayToRowvec(responsesArr, responsesArrSize);


	try
	{
		// run the model function
		regressor.Train(data, responsesVector);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}
	isModelTrained = true;
}

// input:
// output: double
double variance()
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not Trained!");
		return 0.0;
	}
	
	return regressor.Variance();
}

// TODO: remove testing Comment

/*
void testing()
{
	// load the iris data
	mat dataset;
	bool loaded = mlpack::data::Load("iris.csv", dataset);

	// split the data into train and test
	mat trainData = dataset.cols(1, 4);
	trainData.shed_row(trainData.n_rows - 1);
	rowvec trainTarget = trainData.row(trainData.n_rows - 1);
	trainData.shed_row(trainData.n_rows - 1);

	mat testData = dataset.col(dataset.n_cols - 1);
	testData.shed_row(testData.n_rows - 1);
	rowvec testTarget = testData.row(testData.n_rows - 1);
	testData.shed_row(testData.n_rows - 1);


	// init the bayesian linear regressor model


	// train the model
	regressor.Train(trainData, trainTarget);

	// predict with the test data
	rowvec prediction;
	regressor.Predict(testData, prediction);


	// compare test target and prediction
	cout << "Train Data: " << trainData << endl;
	cout << "Train Target: " << trainTarget << endl;
	cout << "Alpha" << alpha() << endl;
	cout << "Beta" << beta() << endl;
}
*/
