
:- module(bayesian_linear_regression, [ blr_initModel/4, 
                                        blr_alpha/1, 
                                        blr_beta/1, 
                                        blr_dataOffset/1, 
                                        blr_dataScale/1, 
                                        blr_omega/1, 
                                        blr_predict/3, 
                                        blr_predictWithStd/4, 
                                        blr_rmse/4, 
                                        blr_train/3, 
                                        blr_variance/1]).

:- load_files(library(str_decl),
                [when(compile_time), if(changed)]).

%% needed for using the array type and for reading from csv
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function and what there inputs and output arguments are


%% --Input--
%%              bool    centerData      => (1)true / (0)false => true,
%%              bool    scaleData       => (1)true / (0)false => false,
%%              int     nIterMax        => 50,
%%              float32 tol             => 1e-4
%%
%% --Output--
%%
%% --Description--
%%              Should be called before all other predicates!
%%              Initiates the Model so now train/5 can be called.
%%              Before predict/5 or predictWitStd/7 can be used train/5 has to be called before
%%
blr_initModel(CenterData, ScaleData, NIterMax, Tol) :-
        NIterMax >= 0,
        Tol > 0,
        initModelI(CenterData, ScaleData, NIterMax, Tol).

foreign(initModel,  c, initModelI(      +integer, 
                                        +integer, 
                                        +integer, +float32)).


%% --Input--
%%
%% --Output--
%%              float32 alpha
%%
%% --Description--
%%              Get the precision (or inverse variance) of the gaussian prior.
%%              train/5 should be called before.
%%
blr_alpha(Alpha) :-
        alphaI(Alpha).

foreign(alpha,  c, alphaI([-float32])).


%% --Input--
%%              float32 beta
%%
%% --Output--
%%
%% --Description--
%%              Get the precision (or inverse variance) beta of the model.
%%              train/5 should be called before.
%%
blr_beta(Beta) :-
        betaI(Beta).

foreign(beta,  c, betaI([-float32])).


%% --Input--
%%
%% --Output--
%%              vec     responsesOffset
%%
%% --Description--
%%              Get the mean vector computed on the features over the training points.
%%
blr_dataOffset(ResponsesList) :-
        dataOffsetI(X, Xsize),
        convert_float_array_to_list(X, Xsize, ResponsesList).

foreign(dataOffset, c, dataOffsetI(-pointer(float_array), -integer)).


%% --Input--
%%
%% --Output--
%%              vec     dataOffset
%%
%% --Description--
%%              Get the vector of standard deviations computed on the features over the training points.
%%
blr_dataScale(DataOffsetList) :-
        dataScaleI(X, Xsize),
        convert_float_array_to_list(X, Xsize, DataOffsetList).

foreign(dataScale, c, dataScaleI(-pointer(float_array), -integer)).


%% --Input--
%%
%% --Output--
%%              vec     omega
%%
%% --Description--
%%              Get the solution vector.
%%
blr_omega(OmegaList) :-
        omegaI(X, Xsize),
        convert_float_array_to_list(X, Xsize, OmegaList).

foreign(omega, c, omegaI(-pointer(float_array), -integer)).


%% --Input--
%%              mat     points
%%
%% --Output--
%%              vec     predictions
%%
%% --Description--
%%              Predict yi for each data point in the given data matrix using the currently-trained Bayesian Ridge model.
%%
blr_predict(PointsList, PointsRows, PredictionsList) :-
        convert_list_to_float_array(PointsList, PointsRows, array(Xsize, Xrows, X)),
        predictI(X, Xsize, Xrows, Y, Ysize),
        convert_float_array_to_list(Y, Ysize, PredictionsList).

foreign(predict,  c, predictI(  +pointer(float_array), +integer, +integer, 
                                -pointer(float_array), -integer)).


%% --Input--
%%              mat     points
%%
%% --Output--
%%              vec     predictions,
%%              vec     std
%%
%% --Description--
%%              Predict yi and the standard deviation of the predictive posterior distribution for each data point in the given data matrix, using the currently-trained Bayesian Ridge estimator.
%%
blr_predictWithStd(PointsList, PointsRows, PredictionsList, STDList) :-
        convert_list_to_float_array(PointsList, PointsRows, array(Xsize, Xrows, X)),
        predictWithStdI(X, Xsize, Xrows, Y, Ysize, Z, Zsize),
        convert_float_array_to_list(Y, Ysize, PredictionsList),
        convert_float_array_to_list(Z, Zsize, STDList).

foreign(predictWithStd,  c, predictWithStdI(    +pointer(float_array), +integer, +integer, 
                                                -pointer(float_array), -integer, 
                                                -pointer(float_array), -integer)).


%% --Input--
%%              mat     data,
%%              vec     responses
%%
%% --Output--
%%              float32 rmse
%%
%% --Description--
%%              Compute the Root Mean Square Error between the predictions returned by the model and the true responses.
%%
blr_rmse(DataList, DataRows, ResponsesList, RMSE) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrows, X)),
        convert_list_to_float_array(ResponsesList, array(Ysize, Y)),
        rmseI(X, Xsize, Xrows, Y, Ysize, RMSE).

foreign(rmse,  c, rmseI(+pointer(float_array), +integer, +integer, 
                        +pointer(float_array), +integer, 
                        [-float32])).


%% --Input--
%%              mat     data,
%%              vec     responses
%%
%% --Output--
%%
%% --Description--
%%              Run BayesianLinearRegression.
%%              The input matrix (like all mlpack matrices) should be column-major each column is an observation and each row is a dimension.
%%
blr_train(DataList, DataRows, ResponsesList) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrows, X)),
        convert_list_to_float_array(ResponsesList, array(Ysize, Y)),
        trainI(X, Xsize, Xrows, Y, Ysize).

foreign(train,  c, trainI(      +pointer(float_array), +integer, +integer, 
                                +pointer(float_array), +integer)).


%% --Input--
%%
%% --Output--
%%              float32 variance
%%
%% --Description--
%%              Get the estimate variance.
%%              train/5 should be called before.
%%
blr_variance(Variance) :-
        varianceI(Variance).

foreign(variance,  c, varianceI([-float32])).

%% Defines what functions should be connected from main.cpp
foreign_resource(bayesian_linear_regression, [  initModel, 
                                                alpha, 
                                                beta, 
                                                dataOffset, 
                                                dataScale, 
                                                omega, 
                                                predict, 
                                                predictWithStd, 
                                                rmse,
                                                train, 
                                                variance]).

:- load_foreign_resource(bayesian_linear_regression).
