#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "decision_tree_glue.h"
#include <mlpack/methods/decision_tree/decision_tree.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::tree;

// Global Variable of the DecisionTree object so it can be accessed from all functions
DecisionTree decisionTreeObj;


// input:   const arma::mat & 	            dataset,
//          const arma::vec &            	labels,
//          const size_t 	                numClasses,
//          const size_t 	                minimumLeafSize = 10,
//          const double 	                minimumGainSplit = 1e-7,
//          const size_t 	                maximumDepth = 0
// output: 
// description: 
void initModel( float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum,
                        float *labelsArr, SP_integer labelsArrSize,
                        SP_integer numClasses, SP_integer minimumLeafSize, double minimumGainSplit, SP_integer maximumDepth)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    // check if labels fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }

    // convert the Prolog array to arma::rowvec
    Row<size_t> labelsVector = convertArrayToVec(labelsArr, labelsArrSize);

    try
    {
        decisionTreeObj = DecisionTree(data, labelsVector, numClasses, minimumLeafSize, minimumGainSplit, maximumDepth);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    
    decisionTreeObj = DecisionTree(data, labelsVector, numClasses, minimumLeafSize, minimumGainSplit, maximumDepth);
}


// input:   const arma::mat & 	dataset,
//          arma::Row< size_t > & 	predictions <-,
//          arma::rowvec & 	probabilities <-
// output: 
// description: 
void classifyPoint( float *pointArr, SP_integer pointArrSize,
                    SP_integer *predict,
                    float **probsArr, SP_integer *probsArrSize)
{
    // convert the Prolog array to arma::rowvec
    rowvec pointVector = convertArrayToRowvec(pointArr, pointArrSize);

    // create the prediction variable
    size_t predictReturn;
    
    // create the ReturnVector
    vec probsReturnVector;
    

    try
    {
        decisionTreeObj.Classify(pointVector, predictReturn, probsReturnVector);
    }
    catch(const std::logic_error& e)
    {
        raisePrologSystemExeption("Given Point has the wrong Length!");
        return;
    }


    // return prediction value
    *predict = predictReturn;

    // return the Vector
    returnVectorInformation(probsReturnVector, probsArr, probsArrSize);
}


// input:   const arma::mat & 	dataset,
//          arma::Row< size_t > & 	predictions <-,
//          arma::rowvec & 	probabilities <-
// output: 
// description: 
void classifyMatrix(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum,
                float **predictArr, SP_integer *predictArrSize,
                float **probsMatArr, SP_integer *probsMatColNum, SP_integer *probsMatRowNum)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    
    // create the ReturnVector
    Row< size_t > predictReturnVector;
    // create the ReturnMat
    mat probsReturnMat;
    

    try
    {
        decisionTreeObj.Classify(data, predictReturnVector, probsReturnMat);
    }
    catch(const std::logic_error& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }

    
    // return the Vector
    returnVectorInformation(predictReturnVector, predictArr, predictArrSize);
    /// return the Matrix
    returnMatrixInformation(probsReturnMat, probsMatArr, probsMatColNum, probsMatRowNum);
    
}


// input:   MatType 	            dataset,
//          LabelsType 	            labels,
//          const size_t 	        numClasses,
//          const size_t 	        minimumLeafSize = 10,
//          const double 	        minimumGainSplit = 1e-7,
//          const size_t 	        maximumDepth = 0,
//          DimensionSelectionType 	dimensionSelector = DimensionSelectionType()
//
// output:  double                  final entropy of decision tree
// description: 
double train(   float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum,
                float *labelsArr, SP_integer labelsArrSize,
                SP_integer numClasses, SP_integer minimumLeafSize, double minimumGainSplit, SP_integer maximumDepth)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    // check if labels fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }

    // convert the Prolog array to arma::rowvec
    Row<size_t> labelsVector = convertArrayToVec(labelsArr, labelsArrSize);

    try
    {
        return decisionTreeObj.Train(data, labelsVector, numClasses, minimumLeafSize, minimumGainSplit, maximumDepth);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return 0.0;
    }
    
    
}
