
:- module(decision_tree, [      decision_tree_initModel/7,
                                decision_tree_classifyPoint/3,
                                decision_tree_classifyMatrix/5,
                                decision_tree_train/8]).

%% requirements of library(struct)
:- load_files(library(str_decl),
        [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function


%% --Input--
%%              mat     dataset,
%%              vec     labels,
%%              int     numClasses,
%%              int     minimumLeafSize         => 10,
%%              float32 minimumGainSplit        => 1e-7,
%%              int     maximumDepth            => 0,
%%
%% --Output--
%%
%% --Description--
%%              Construct the decision tree on the given data and labels, assuming that the data is all of the numeric type.
%%              Setting minimumLeafSize and minimumGainSplit too small may cause the tree to overfit, but setting them too large may cause it to underfit.
%%
decision_tree_initModel(DataList, DataRows, LabelsList, NumClasses, MinimumLeafSize, MinimumGainSplit, MaximumDepth) :-
        NumClasses >= 0,
        MinimumLeafSize > 0,
        MinimumGainSplit > 0.0,
        MinimumGainSplit < 1.0,
        MaximumDepth >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        initModelI(X, Xsize, Xrownum, Y, Ysize, NumClasses, MinimumLeafSize, MinimumGainSplit, MaximumDepth).

foreign(initModel, c, initModelI(       +pointer(float_array), +integer, +integer,
                                        +pointer(float_array), +integer,
                                        +integer, +integer, +float32, +integer)).


%% --Input--
%%              vec     point
%%
%% --Output--
%%              int     prediction,
%%              vec     probabilities
%%
%% --Description--
%%              Classify the given point and also return estimates of the probability for each class in the given vector.
%%
decision_tree_classifyPoint(DataList, Prediction, AssignList) :-
        convert_list_to_float_array(DataList, array(Xsize, X)),
        classifyPointI(X, Xsize, Prediction, Y, Ysize),
        convert_float_array_to_list(Y, Ysize, AssignList).

foreign(classifyPoint, c, classifyPointI(       +pointer(float_array), +integer,
                                                -integer,
                                                -pointer(float_array), -integer)).


%% --Input--
%%              mat     data
%%
%% --Output--
%%              vec     predictions,
%%              mat     probabilities
%%
%% --Description--
%%              Classify the given points and also return estimates of the probabilities for each class in the given matrix.
%%
decision_tree_classifyMatrix(DataList, DataRows, PredictionList, ProbsList, ZCols) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrows, X)),
        classifyMatrixI(X, Xsize, Xrows, Y, Ysize, Z, ZCols, ZRows),
        convert_float_array_to_list(Y, Ysize, PredictionList),
        convert_float_array_to_2d_list(Z, ZCols, ZRows, ProbsList).

foreign(classifyMatrix, c, classifyMatrixI(     +pointer(float_array), +integer, +integer,
                                                -pointer(float_array), -integer,
                                                -pointer(float_array), -integer, -integer)).


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses,
%%              int     minimumLeafSize         => 10,
%%              float32 minimumGainSplit        => 1e-7,
%%              int     maximumDepth            => 0,
%%
%% --Output--
%%              float32 final entropy of decision tree
%%
%% --Description--
%%              Train the decision tree on the given data, assuming that all dimensions are numeric.
%%              This will overwrite the given model. Setting minimumLeafSize and minimumGainSplit too small may cause the tree to overfit, but setting them too large may cause it to underfit.
%%
decision_tree_train(DataList, DataRows, LabelsList, NumClasses, MinimumLeafSize, MinimumGainSplit, MaximumDepth, Entropy) :-
        NumClasses >= 0,
        MinimumLeafSize > 0,
        MinimumGainSplit > 0.0,
        MinimumGainSplit < 1.0,
        MaximumDepth >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        trainI(X, Xsize, Xrownum, Y, Ysize, NumClasses, MinimumLeafSize, MinimumGainSplit, MaximumDepth, Entropy).

foreign(train, c, trainI(       +pointer(float_array), +integer, +integer,
                                +pointer(float_array), +integer,
                                +integer, +integer, +float32, +integer,
                                [-float32])).


%% Defines the functions that get connected from main.cpp
foreign_resource(decision_tree, [       initModel,
                                        classifyPoint,
                                        classifyMatrix,
                                        train]).

:- load_foreign_resource(decision_tree).
