#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "fastmks_glue.h"
#include <mlpack/methods/fastmks/fastmks_model.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::fastmks;
using namespace mlpack::kernel;

// Global Variable of the FastMKSModel object so it can be accessed from all functions
FastMKSModel fastMKSModel;
bool isModelTrained = false;


// input:   util::Timers & 	timers,
//          arma::mat && 	referenceData,
//          TKernelType & 	kernel,
//			double 			degree,
//			double 			offset,
//			double 			bandwidth,
//			double 			scale
//          const bool 	    singleMode,
//          const bool 	    naive,
//          const double 	base
// output: 
// description: 
//          Initializes the model on the given reference set.
//
void initModel(float *referenceMatArr, SP_integer referenceMatSize, SP_integer referenceMatRowNum, 
                char const *kernel, 
                double degree, double offset, double bandwidth, double scale, 
                SP_integer singleMode, 
                SP_integer naive, 
                double base)
{
    // convert the Prolog array to arma::mat
    mat reference = convertArrayToMat(referenceMatArr, referenceMatSize, referenceMatRowNum);

    if (strcmp(kernel, "linear") == 0)        
    {
        fastMKSModel = FastMKSModel(0);
        LinearKernel lk;
        fastMKSModel.BuildModel(move(reference), lk, (singleMode == 1), (naive == 1), base);
        isModelTrained = true;
    }
    else if (strcmp(kernel, "polynomial") == 0)
    {
        PolynomialKernel pk(degree, offset);
        fastMKSModel.KernelType() = FastMKSModel::POLYNOMIAL_KERNEL;
        fastMKSModel.BuildModel(move(reference), pk, (singleMode == 1), (naive == 1), base);
        isModelTrained = true;
    }
    else if (strcmp(kernel, "cosine") == 0)
    {
        CosineDistance cd;
        fastMKSModel.KernelType() = FastMKSModel::COSINE_DISTANCE;
        fastMKSModel.BuildModel(move(reference), cd, (singleMode == 1), (naive == 1), base);
        isModelTrained = true;
    }
    else if (strcmp(kernel, "gaussian") == 0)
    {
        GaussianKernel gk(bandwidth);
        fastMKSModel.KernelType() = FastMKSModel::GAUSSIAN_KERNEL;
        fastMKSModel.BuildModel(move(reference), gk, (singleMode == 1), (naive == 1), base);
        isModelTrained = true;
    }
    else if (strcmp(kernel, "epanechnikov") == 0)
    {
        EpanechnikovKernel ek(bandwidth);
        fastMKSModel.KernelType() = FastMKSModel::EPANECHNIKOV_KERNEL;
        fastMKSModel.BuildModel(move(reference), ek, (singleMode == 1), (naive == 1), base);
        isModelTrained = true;
    }
    else if (strcmp(kernel, "triangular") == 0)
    {
        TriangularKernel tk(bandwidth);
        fastMKSModel.KernelType() = FastMKSModel::TRIANGULAR_KERNEL;
        fastMKSModel.BuildModel(move(reference), tk, (singleMode == 1), (naive == 1), base);
        isModelTrained = true;
    }
    else if (strcmp(kernel, "hyptan") == 0)
    {
        
        HyperbolicTangentKernel htk(scale, offset);
        fastMKSModel.KernelType() = FastMKSModel::HYPTAN_KERNEL;
        fastMKSModel.BuildModel(move(reference), htk, (singleMode == 1), (naive == 1), base);
        isModelTrained = true;
    }
    else
    {
        raisePrologDomainExeption(kernel, 2, "The given kernel is unkown!", "initModel");
    }

}


// input:   util::Timers & 	        timers,
//          const arma::mat & 	    querySet,
//          const size_t 	        k,
//          arma::Mat< size_t > & 	indices <-,
//          arma::mat & 	        kernels <-,
//          const double 	        base
// output: 
// description: 
//          Search with a different query set.
//
void searchWithQuery(float *querySetMatArr, SP_integer querySetMatSize, SP_integer querySetMatRowNum, 
                        SP_integer k, 
                        float **indicesMatArr, SP_integer *indicesMatColNum, SP_integer *indicesMatRowNum, 
                        float **kernelsMatArr, SP_integer *kernelsMatColNum, SP_integer *kernelsMatRowNum, 
                        double base)
{
    if (!isModelTrained)
    {
        raisePrologSystemExeption("The Model hasnt been trained yet!");
        return;
    }
    
    // convert the Prolog array to arma::mat
    mat querySet = convertArrayToMat(querySetMatArr, querySetMatSize, querySetMatRowNum);
    
    // create the ReturnMats
    arma::Mat< size_t > indicesReturnMat;
    mat kernelsReturnMat;

    
    try
    {
        fastMKSModel.Search(querySet, k, indicesReturnMat, kernelsReturnMat, base);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
    }

    
    // return the Matrices
    returnMatrixInformation(indicesReturnMat, indicesMatArr, indicesMatColNum, indicesMatRowNum);
    returnMatrixInformation(kernelsReturnMat, kernelsMatArr, kernelsMatColNum, kernelsMatRowNum);
}


// input:   util::Timers & 	        timers,
//          const size_t 	        k,
//          arma::Mat< size_t > & 	indices <-,
//          arma::mat & 	        kernels <-
// output: 
// description: 
//          Search with the reference set as the query set.
//
void searchNoQuery(SP_integer k, 
                    float **indicesMatArr, SP_integer *indicesMatColNum, SP_integer *indicesMatRowNum, 
                    float **kernelsMatArr, SP_integer *kernelsMatColNum, SP_integer *kernelsMatRowNum)
{
    if (!isModelTrained)
    {
        raisePrologSystemExeption("The Model hasnt been trained yet!");
        return;
    }
    // create the ReturnMats
    arma::Mat< size_t > indicesReturnMat;
    arma::mat kernelsReturnMat;
    
    
    try
    {
        fastMKSModel.Search((size_t)k, indicesReturnMat, kernelsReturnMat);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }


    returnMatrixInformation(indicesReturnMat, indicesMatArr, indicesMatColNum, indicesMatRowNum);
    returnMatrixInformation(kernelsReturnMat, kernelsMatArr, kernelsMatColNum, kernelsMatRowNum);
}
