
:- module(hoeffding_tree, [     hoeffding_tree_initAndBuildModel/12,
                                hoeffding_tree_classify/4,
                                hoeffding_tree_train/4]).

%% requirements of library(struct)
:- load_files(library(str_decl),
        [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function


%% --Input--
%%              string  treeType                => "gini_hoeffding", "gini_binary", "info_hoeffding", "info_binary",
%%              mat     dataset,
%%              vec     labels,
%%              int     numClasses,
%%              bool    batchTrainig            => (1)true / (0)false,
%%              float32 successProbability,
%%              int     maxSamples,
%%              int     checkInterval,
%%              int     minSamples,
%%              int     bins,
%%              int     observationsBeforeBinning
%%
%% --Output--
%%
%% --Description--
%%              Construct the Hoeffding tree with the given parameters and given training data.
%%              The tree may be trained either in batch mode (which looks at all points before splitting, and propagates these points to the created children for further training), or in streaming mode, where each point is only considered once. (In general, batch mode will give better-performing trees, but will have higher memory and runtime costs for the same dataset.)
%%
hoeffding_tree_initAndBuildModel(TreeType, DataList, DataRows, LabelsList, NumClasses, BatchTraining, SuccessProbability, MaxSamples, CheckInterval, MinSamples, Bins, ObservationsBeforeBinning) :-
        NumClasses >= 0,
        SuccessProbability >= 0,
        SuccessProbability =< 1,
        MaxSamples >= 0,
        CheckInterval > 0,
        MinSamples >= 0,
        Bins >= 0,
        ObservationsBeforeBinning >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        initAndBuildModelI(TreeType, X, Xsize, Xrownum, Y, Ysize, NumClasses, BatchTraining, SuccessProbability, MaxSamples, CheckInterval, MinSamples, Bins, ObservationsBeforeBinning).

foreign(initAndBuildModel, c, initAndBuildModelI(       +string,
                                                        +pointer(float_array), +integer, +integer,
                                                        +pointer(float_array), +integer,
                                                        +integer, +integer, +float32, +integer, +integer, +integer, +integer, +integer)).



%% --Input--
%%              mat     data
%%
%% --Output--
%%              vec     predictions,
%%              vec     probabilities
%%
%% --Description--
%%              Classify the given points, using this node and the entire (sub)tree beneath it.
%%              The predicted labels for each point are returned, as well as an estimate of the probability that the prediction is correct for each point. This estimate is simply the MajorityProbability for the leaf that each point bins to.
%%
hoeffding_tree_classify(TestList, TestRows, PredicList, ProbsList) :-
        convert_list_to_float_array(TestList, TestRows, array(Xsize, Xrownum, X)),
        classifyI(X, Xsize, Xrownum, Y, Ysize, Z, Zsize),
        convert_float_array_to_list(Y, Ysize, PredicList),
        convert_float_array_to_list(Z, Zsize, ProbsList).

foreign(classify, c, classifyI( +pointer(float_array), +integer, +integer,
                                -pointer(float_array), -integer,
                                -pointer(float_array), -integer)).



%% --Input--
%%              mat     data,
%%              vec     labels,
%%              bool    batchTrainig            => (1)true / (0)false
%%
%% --Output--
%%
%% --Description--
%%              Train in streaming mode on the given dataset.
%%              This takes one pass. Be sure that initAndBuildModel/14 has been called first!
%%
hoeffding_tree_train(DataList, DataRows, LabelsList, BatchTraining) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        trainI(X, Xsize, Xrownum, Y, Ysize, BatchTraining).

foreign(train, c, trainI(       +pointer(float_array), +integer, +integer,
                                +pointer(float_array), +integer,
                                +integer)).


%% Defines the functions that get connected from main.cpp
foreign_resource(hoeffding_tree, [      initAndBuildModel,
                                        classify,
                                        train]).

:- load_foreign_resource(hoeffding_tree).
