#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "kde_glue.h"
#include <mlpack/methods/kde/kde_model.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::kde;

// Global Variable of the KDEModel object so it can be accessed from all functions
KDEModel kdeModel;


// input:   const double 	    bandwidth = 1.0,
//          const double 	    relError = KDEDefaultParams::relError,
//          const double 	    absError = KDEDefaultParams::absError,
//          const KernelTypes 	kernelType = KernelTypes::GAUSSIAN_KERNEL,
//          const TreeTypes 	treeType = TreeTypes::KD_TREE,
//          const bool 	        monteCarlo = KDEDefaultParams::mode,
//          const double 	    mcProb = KDEDefaultParams::mcProb,
//          const size_t 	    initialSampleSize = KDEDefaultParams::initialSampleSize,
//          const double 	    mcEntryCoef = KDEDefaultParams::mcEntryCoef,
//          const double 	    mcBreakCoef = KDEDefaultParams::mcBreakCoef,
//          arma::mat && 	    referenceSet,
// output: 
// description: 
//          Build the KDE model with the given parameters and then trains it with the given reference data.
//
void initAndBuildModel(double bandwidth, double relError, double absError, 
                        char const *kernelType, char const *treeType, char const *algorithm,
                        SP_integer monteCarlo, double mcProb, SP_integer initialSampleSize, double mcEntryCoef, double mcBreakCoef,
                        float *referenceMatArr, SP_integer referenceMatSize, SP_integer referenceMatRowNum)
{
    // convert the Prolog array to arma::mat
    mat reference = convertArrayToMat(referenceMatArr, referenceMatSize, referenceMatRowNum);

    // select the kernelType
    mlpack::kde::KDEModel::KernelTypes kernelT;

    if (strcmp(kernelType, "gaussian") == 0)
    {
        kernelT = KDEModel::GAUSSIAN_KERNEL;
    }
    else if (strcmp(kernelType, "epanechnikov") == 0)
    {
        kernelT = KDEModel::EPANECHNIKOV_KERNEL;
    }
    else if (strcmp(kernelType, "laplacian") == 0)
    {
        kernelT = KDEModel::LAPLACIAN_KERNEL;
    }
    else if (strcmp(kernelType, "spherical") == 0)
    {
        kernelT = KDEModel::SPHERICAL_KERNEL;
    }
    else if (strcmp(kernelType, "triangular") == 0)
    {
        kernelT = KDEModel::TRIANGULAR_KERNEL;
    }
    else
    {
        raisePrologDomainExeption(kernelType, 4, "The given KernelType is unknown!", "initAndBuildModel");
        return;
    }

    // select the treeType
    mlpack::kde::KDEModel::TreeTypes treeT;

    if (strcmp(treeType, "kd_tree") == 0)
    {
        treeT = KDEModel::KD_TREE;
    }
    else if (strcmp(treeType, "ball_tree") == 0)
    {
        treeT = KDEModel::BALL_TREE;
    }
    else if (strcmp(treeType, "cover_tree") == 0)
    {
        treeT = KDEModel::COVER_TREE;
    }
    else if (strcmp(treeType, "octree") == 0)
    {
        treeT = KDEModel::OCTREE;
    }
    else if (strcmp(treeType, "r_tree") == 0)
    {
        treeT = KDEModel::R_TREE;
    }
    else
    {
        raisePrologDomainExeption(treeType, 5, "The given TreeType is unknown!", "initAndBuildModel");
        return;
    }
    
    
    try
    {
        kdeModel = KDEModel(bandwidth, relError, absError, kernelT, treeT, (monteCarlo == 1), mcProb, initialSampleSize, mcEntryCoef, mcBreakCoef);
        kdeModel.BuildModel(move(reference));
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }


    if (strcmp(algorithm, "dual_tree") == 0)
    {
        KDEMode& mode = kdeModel.Mode();
        mode = KDEMode::DUAL_TREE_MODE;
    }
    else if (strcmp(algorithm, "single_tree") == 0)
    {
        KDEMode& mode = kdeModel.Mode();
        mode = KDEMode::SINGLE_TREE_MODE;
    }
    else
    {
        raisePrologDomainExeption(algorithm, 6, "The given Algorithm is unknown!", "initAndBuildModel");
        return;
    }
}


// input:   arma::mat && 	    querySet,
//          arma::vec & 	    estimations <-
// output: 
// description: 
//          Perform kernel density estimation on the given query set.
//
void evaluateWithQuery(float *querySetMatArr, SP_integer querySetMatSize, SP_integer querySetMatRowNum, float **estimationsArr, SP_integer *estimationsArrSize)
{
    // convert the Prolog array to arma::mat
    mat querySet = convertArrayToMat(querySetMatArr, querySetMatSize, querySetMatRowNum);

    // create the ReturnVector
    vec estimationsReturnVector;


    try
    {
        kdeModel.Evaluate(move(querySet), estimationsReturnVector);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    
    

    // return the Vector
    returnVectorInformation(estimationsReturnVector, estimationsArr, estimationsArrSize);
}


// input:   arma::vec & 	    estimations <-
// output: 
// description: 
//          Perform kernel density estimation on the reference set.
//
void evaluateNoQuery(float **estimationsArr, SP_integer *estimationsArrSize)
{
    // create the ReturnVector
    vec estimationsReturnVector;


    try
    {
        kdeModel.Evaluate(estimationsReturnVector);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    
    
    // return the Vector
    returnVectorInformation(estimationsReturnVector, estimationsArr, estimationsArrSize);
}