#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "kfn_glue.h"
#include <mlpack/methods/neighbor_search/ns_model.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::neighbor;

// Global Variable of the NSModel object so it can be accessed from all functions
NSModel<FurthestNS> nsModelObj;
int trainedDimension = 0;


// input:   TreeTypes 	                treeType = TreeTypes::KD_TREE,
//          bool 	                    randomBasis = false,
//          util::Timers & 	            timers,
//          arma::mat && 	            referenceSet,
//          const NeighborSearchMode 	searchMode,
//          const double 	            epsilon = 0
// output: 
// description: 
//          Initialize the Model and build it.
//
void initAndBuildModel(char const *treeType, char const *algorithm,
                        SP_integer randomBasis, SP_integer leafSize, double epsilon,
                        float *referenceMatArr, SP_integer referenceMatSize, SP_integer referenceMatRowNum)
{
    // convert the Prolog array to arma::mat
    mat reference = convertArrayToMat(referenceMatArr, referenceMatSize, referenceMatRowNum);


    // used a good part of how it was used in the .main file
    // select the treeType
    NSModel<FurthestNS>::TreeTypes tree = NSModel<FurthestNS>::KD_TREE;

    if (strcmp(treeType, "kd") == 0)
        tree = NSModel<FurthestNS>::KD_TREE;
    else if (strcmp(treeType, "cover") == 0)
        tree = NSModel<FurthestNS>::COVER_TREE;
    else if (strcmp(treeType, "r") == 0)
        tree = NSModel<FurthestNS>::R_TREE;
    else if (strcmp(treeType, "r_star") == 0)
        tree = NSModel<FurthestNS>::R_STAR_TREE;
    else if (strcmp(treeType, "ball") == 0)
        tree = NSModel<FurthestNS>::BALL_TREE;
    else if (strcmp(treeType, "x") == 0)
        tree = NSModel<FurthestNS>::X_TREE;
    else if (strcmp(treeType, "hilbert_r") == 0)
        tree = NSModel<FurthestNS>::HILBERT_R_TREE;
    else if (strcmp(treeType, "r_plus") == 0)
        tree = NSModel<FurthestNS>::R_PLUS_TREE;
    else if (strcmp(treeType, "r_plus_plus") == 0)
        tree = NSModel<FurthestNS>::R_PLUS_PLUS_TREE;
    else if (strcmp(treeType, "vp") == 0)
        tree = NSModel<FurthestNS>::VP_TREE;
    else if (strcmp(treeType, "rp") == 0)
        tree = NSModel<FurthestNS>::RP_TREE;
    else if (strcmp(treeType, "max_rp") == 0)
        tree = NSModel<FurthestNS>::MAX_RP_TREE;
    else if (strcmp(treeType, "ub") == 0)
        tree = NSModel<FurthestNS>::UB_TREE;
    else if (strcmp(treeType, "oct") == 0)
        tree = NSModel<FurthestNS>::OCTREE;
    else
        raisePrologDomainExeption(treeType, 1, "The given TreeType is unknown!", "initAndBuildModel");


    //  select the searchMode
    NeighborSearchMode searchMode = DUAL_TREE_MODE;

    if (strcmp(algorithm, "naive") == 0)
        searchMode = NAIVE_MODE;
    else if (strcmp(algorithm, "single_tree") == 0)
        searchMode = SINGLE_TREE_MODE;
    else if (strcmp(algorithm, "dual_tree") == 0)
        searchMode = DUAL_TREE_MODE;
    else if (strcmp(algorithm, "greedy") == 0)
        searchMode = GREEDY_SINGLE_TREE_MODE;
    else
        raisePrologDomainExeption(algorithm, 2, "The given SearchMode is unknown!", "initAndBuildModel");

    
    nsModelObj.TreeType() = tree;
    nsModelObj.RandomBasis() = (randomBasis == 1);

    
    trainedDimension = reference.n_rows;
    nsModelObj.BuildModel(std::move(reference), size_t(leafSize), searchMode, epsilon);
}


// input:   	util::Timers & 	        timers,
//              arma::mat && 	        querySet,
//              const size_t 	        k,
//              arma::Mat< size_t > & 	neighbors <-,
//              arma::mat & 	        distances <-
// output: 
// description: 
//              Perform neighbor search on the queryset.
//
void searchWithQuery(   float *queryMatArr, SP_integer queryMatSize, SP_integer queryMatRowNum,
                        SP_integer k,
                        float **neighborsMatArr, SP_integer *neighborsMatColNum, SP_integer *neighborsMatRowNum,
                        float **distancesMatArr, SP_integer *distancesMatColNum, SP_integer *distancesMatRowNum)
{
    if (trainedDimension != queryMatRowNum)
    {
        std::ostringstream stream;
        stream << "Queryset has Dim(" << queryMatRowNum << ") but the Referenceset has Dim(" << trainedDimension << ")";
        raisePrologSystemExeption(stream.str().c_str());
        return;
    }
    
    // convert the Prolog array to arma::mat
    mat query = convertArrayToMat(queryMatArr, queryMatSize, queryMatRowNum);

    // create the ReturnMat
    Mat< size_t > neighborsReturnMat;
    mat distancesReturnMat;
    
    try
    {
        nsModelObj.Search(move(query), (size_t)k, neighborsReturnMat, distancesReturnMat);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
    }


    // return the Matrices
    returnMatrixInformation(neighborsReturnMat, neighborsMatArr, neighborsMatColNum, neighborsMatRowNum);
    returnMatrixInformation(distancesReturnMat, distancesMatArr, distancesMatColNum, distancesMatRowNum);
}


// input:   	util::Timers & 	        timers,
//              const size_t 	        k,
//              arma::Mat< size_t > & 	neighbors <-,
//              arma::mat & 	        distances <-
// output: 
// description: 
//              Perform monochromatic neighbor search.
//
void searchNoQuery(     SP_integer k,
                        float **neighborsMatArr, SP_integer *neighborsMatColNum, SP_integer *neighborsMatRowNum,
                        float **distancesMatArr, SP_integer *distancesMatColNum, SP_integer *distancesMatRowNum)
{
    // create the ReturnMats
    Mat< size_t > neighborsReturnMat;
    mat distancesReturnMat;
    
    try
    {
        nsModelObj.Search((size_t)k, neighborsReturnMat, distancesReturnMat);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
    }


    // return the Matrices
    returnMatrixInformation(neighborsReturnMat, neighborsMatArr, neighborsMatColNum, neighborsMatRowNum);
    returnMatrixInformation(distancesReturnMat, distancesMatArr, distancesMatColNum, distancesMatRowNum);
}
