
:- module(lars, [       lars_initAndTrainModel/9, 
                        lars_activeSet/1, 
                        lars_beta/1, 
                        lars_betaPath/2, 
                        lars_computeError/5, 
                        lars_lambdaPath/1, 
                        lars_matUtriCholFactor/2, 
                        lars_predict/4]).

:- load_files(library(str_decl),
                [when(compile_time), if(changed)]).

%% needed for using the array type and for reading from csv
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function and what there inputs and output arguments are


%% --Input--
%%              bool    useCholesky     => (1)true / (0)false =>false,
%%              float32 lambda1         => 0.0,
%%              float32 lambda2         => 0.0,
%%              float32 tolerance       => 1e-16
%%              mat     data,
%%              vec     responses,
%%              bool    rowMajor        => (1)true / (0)false => true
%%
%% --Output--
%%              float32 error
%%
%% --Description--
%%              Only initialize the LARS model.
%%
lars_initAndTrainModel(UseCholesky, Lambda1, Lambda2, Tolerance, DataList, DataPointsDim, ResponsesList, RowMajor, Error) :-
        Tolerance > 0, 
        convert_list_to_float_array(DataList, DataPointsDim, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(ResponsesList, array(Ysize, Y)),
        initAndTrainModelI(UseCholesky, Lambda1, Lambda2, Tolerance, X, Xsize, Xrownum, Y, Ysize, RowMajor, Error).

foreign(initAndTrainModel, c, initAndTrainModelI(     +integer, 
                                                                +float32, +float32, +float32,
                                                                +pointer(float_array), +integer, +integer,
                                                                +pointer(float_array), +integer, 
                                                                +integer, -float32)).



%% --Input--
%%
%% --Output--
%%              vec     activeSet
%%
%% --Description--
%%              Get the set of active dimensions
%%
lars_activeSet(ActiveSetList) :-
        activeSetI(Y, Ysize),
        convert_float_array_to_list(Y, Ysize, ActiveSetList).

foreign(activeSet, c, activeSetI(-pointer(float_array), -integer)).


%% --Input--
%%
%% --Output--
%%              vec     beta
%%
%% --Description--
%%              Get the solution coefficients.
%%
lars_beta(BetaList) :-
        betaI(Y, Ysize),
        convert_float_array_to_list(Y, Ysize, BetaList).

foreign(beta, c, betaI(-pointer(float_array), -integer)).


%% --Input--
%%
%% --Output--
%%              mat     betaPath
%%
%% --Description--
%%              Get the set of coefficients after each iteration. The solution is the last element.
%%
lars_betaPath(BetaList, XCols) :-
        betaPathI(X, XCols, XRows),
        convert_float_array_to_2d_list(X, XCols, XRows, BetaList).

foreign(betaPath, c, betaPathI(-pointer(float_array), -integer, -integer)).


%% --Input--
%%              mat     matX,
%%              vec     y,
%%              bool    rowMajor        => (1)true / (0)false =>false
%%
%% --Output--
%%              float32 error
%%
%% --Description--
%%              Compute cost error of the given data matrix using the currently-trained LARS model.Only ||y-beta*X||2 is used to calculate cost error.
%%
lars_computeError(DataList, DataRows, ResponsesList, RowMajor, Error) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(ResponsesList, array(Ysize, Y)),
        computeErrorI(X, Xsize, Xrownum, Y, Ysize, RowMajor, Error).

foreign(computeError, c, computeErrorI( +pointer(float_array), +integer, +integer, 
                                        +pointer(float_array), +integer, 
                                        +integer, 
                                        [-float32])).


%% --Input--
%%
%% --Output--
%%              vec     lambdaPath
%%
%% --Description--
%%              Get the set of values for lambda1 after each iteration; the solution is the last element.
%%
lars_lambdaPath(LambdaPathList) :-
        lambdaPathI(Y, Ysize),
        convert_float_array_to_list(Y, Ysize, LambdaPathList).

foreign(lambdaPath, c, lambdaPathI(-pointer(float_array), -integer)).


%% --Input--
%%
%% --Output--
%%              mat     factor
%%
%% --Description--
%%              Get the upper triangular cholesky factor.
%%
lars_matUtriCholFactor(FactorList, XCols) :-
        matUtriCholFactorI(X, XCols, XRows),
        convert_float_array_to_2d_list(X, XCols, XRows, FactorList).

foreign(matUtriCholFactor, c, matUtriCholFactorI(-pointer(float_array), -integer, -integer)).


%% --Input--
%%              mat     points,
%%              bool    rowMajor        => (1)true / (0)false => false
%%
%% --Output--
%%              vec     predictions
%%
%% --Description--
%%              Predict y_i for each data point in the given data matrix using the currently-trained LARS model.
%%
lars_predict(PointsList, PointsRows, PredicList, RowMajor) :-
        convert_list_to_float_array(PointsList, PointsRows, array(Xsize, Xrownum, X)),
        predictI(X, Xsize, Xrownum, Y, Ysize, RowMajor),
        convert_float_array_to_list(Y, Ysize, PredicList).

foreign(predict, c, predictI(   +pointer(float_array), +integer, +integer, 
                                -pointer(float_array), -integer, 
                                +integer)).


%% Defines what functions should be connected from main.cpp
foreign_resource(lars, [initAndTrainModel, 
                        activeSet, 
                        beta, 
                        betaPath, 
                        computeError, 
                        lambdaPath, 
                        matUtriCholFactor, 
                        predict]).

:- load_foreign_resource(lars).
