#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "linear_SVM_glue.h"
#include <mlpack/methods/linear_svm/linear_svm.hpp>
#include <mlpack/core.hpp>


// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::svm;

// Global Variable of the LinearSVM object so it can be accessed from all functions
LinearSVM<> linearSVM;
bool isModelTrained = false;


// input: 	const MatType & 				data,
//			const arma::Row< size_t > & 	labels,
//			const size_t 					numClasses = 2,
//			const double 					lambda = 0.0001,
//			const double 					delta = 1.0,
//			const bool 						fitIntercept = false,
//			OptimizerType 					optimizer = OptimizerType()
// output: 
// description: 
//			Initializes the linear_svm model with the given data and trains it.
//
void initModelWithTrain(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
						float *labelsArr, SP_integer labelsArrSize, 
						SP_integer numClasses, double lambda, double delta, 
						SP_integer fitIntercept, 
						char const *optimizer)
{
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if labels fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }

	// convert the Prolog array to arma::rowvec
	arma::Row<size_t> labels = convertArrayToVec(labelsArr, labelsArrSize);
	
	linearSVM = LinearSVM<>(numClasses, lambda, delta, (fitIntercept == 1));

	if (strcmp(optimizer, "lbfgs") == 0)
	{
		linearSVM.Train<ens::L_BFGS>(data, labels, numClasses);
		isModelTrained = true;
	}
	else if (strcmp(optimizer, "psgd") == 0)
	{
		linearSVM.Train<ens::ParallelSGD<>>(data, labels, numClasses, ens::ParallelSGD(100, 4));
		isModelTrained = true;
	}
	else
	{
		raisePrologDomainExeption(optimizer, 8, "The given Optimizer is unkown!", "initModelWithTrain");
		return;
	}
}


// input:	const size_t 	numClasses = 0,
//			const double 	lambda = 0.0001,
//			const double 	delta = 1.0,
//			const bool 		fitIntercept = false
// output: 
// description: 
//			Initializes the linear_svm model with the given data but doesnt train it.
//
void initModelNoTrain(SP_integer numClasses, double lambda, double delta, SP_integer fitIntercept)
{
	linearSVM = LinearSVM<>(numClasses, lambda, delta, true);

		isModelTrained = false;
}


// input: 	const MatType & 		data,
//			arma::Row< size_t > & 	labels,
//			arma::mat & 			scores
// output: 
// description: 
//			Classify the given points, returning class scores and predicted class label for each point.
//
void classify(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
				float **labelsArr, SP_integer *labelsArrSize, 
				float **scoresMatArr, SP_integer *scoresMatColNum, SP_integer *scoresMatRowNum)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not trained!");
		return;
	}
	
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if data has the correct feature size
	if (data.n_rows != linearSVM.FeatureSize())
	{
		raisePrologSystemExeption("The given Datapoints Dimension is != the trained Datapoints Dimension!");
		return;
	}

	// get the ReturnVector
	arma::Row<size_t> labelsReturnVector;
	// get the ReturnMat
	mat scoresReturnMat;

	try
	{
		linearSVM.Classify(data, labelsReturnVector, scoresReturnMat);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}


	// return the Vector
	returnVectorInformation(labelsReturnVector, labelsArr, labelsArrSize);
	// return the Matrix
	returnMatrixInformation(scoresReturnMat, scoresMatArr, scoresMatColNum, scoresMatRowNum);
}


// input: 	const VecType & 	point
//
// output: 	size_t 				point label
// description: 
//			Classify the given point.
//
SP_integer classifyPoint(float *pointArr, SP_integer pointArrSize)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not trained!");
		return 0;
	}
	// convert the Prolog array to arma::rowvec
	vec pointVector = convertArrayToColvec(pointArr, pointArrSize);
	// check if data has the correct feature size
	if (pointArrSize != linearSVM.FeatureSize())
	{
		raisePrologSystemExeption("The given Datapoint Dimension is != the trained Datapoints Dimension!");
		return 0;
	}

	try
	{
		return linearSVM.Classify<vec>(pointVector);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return 0;
	}
}


// input: 	const MatType & 				testData,
//			const arma::Row< size_t > & 	testLabels
//
// output: 	double 							accuracy
// description: 
//			Computes accuracy of the learned model given the feature data and the labels associated with each data point.
//
double computeAccuracy(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, float *labelsArr, SP_integer labelsArrSize)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not trained!");
		return 0.0;
	}
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if data has the correct feature size
	if (data.n_rows != linearSVM.FeatureSize())
	{
		raisePrologSystemExeption("The given Datapoints Dimension is != the trained Datapoints Dimension!");
		return 0.0;
	}
	// check if labels fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }

	// convert the Prolog array to arma::rowvec
	arma::Row<size_t> labels = convertArrayToVec(labelsArr, labelsArrSize);


	try
	{
		return linearSVM.ComputeAccuracy(data, labels);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return 0.0;
	}
}


// input: 	const MatType & 				data,
//			const arma::Row< size_t > & 	labels,
//			const size_t 					numClasses = 2,
//			OptimizerType 					optimizer = OptimizerType()
//
// output: 	double 							Object value of final point
// description: 
//			Train the Linear_svm model with the given training data.
//
double train(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, float *labelsArr, SP_integer labelsArrSize, SP_integer numClasses, char const *optimizer)
{
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if labels fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }

	// convert the Prolog array to arma::rowvec
	arma::Row<size_t> labels = convertArrayToVec(labelsArr, labelsArrSize);

	try
	{
		if (strcmp(optimizer, "lbfgs") == 0)
		{
			isModelTrained = true;
			return linearSVM.Train<ens::L_BFGS>(data, labels, numClasses);
		}
		else if (strcmp(optimizer, "psgd") == 0)
		{
			isModelTrained = true;
			return linearSVM.Train<ens::ParallelSGD<>>(data, labels, numClasses, ens::ParallelSGD(100, 4));
		}
		else
		{
			raisePrologDomainExeption(optimizer, 5, "The given Optimizer is unkown!", "train");
			return 0.0;
		}
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return 0.0;
	}
}
