
:- module(linear_SVM, [ linear_SVM_initModelWithTrain/8,
                        linear_SVM_initModelNoTrain/4,
                        linear_SVM_classify/5,
                        linear_SVM_classifyPoint/2,
                        linear_SVM_computeAccuracy/4,
                        linear_SVM_train/6]).

%% requirements of library(struct)
:- load_files(library(str_decl),
              [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses      => 2,
%%              float32 lambda          => 0.0001,
%%              float32 delta           => 1.0,
%%              bool    fitIntercept    => (1)true / (0)false => false,
%%              string  optimizer       => "lbfgs", "psgd" => lbfgs
%%
%% --Output--
%%
%% --Description--
%%              Initializes the linear_svm model with the given data and trains it.
%%
linear_SVM_initModelWithTrain(DataList, DataRows, LabelsList, NumClasses, Lambda, Delta, FitIntercept, Optimizer) :-
        NumClasses >= 0,
        Lambda >= 0.0,
        Delta >= 0.0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        initModelWithTrainI(X, Xsize, Xrownum, Y, Ysize, NumClasses, Lambda, Delta, FitIntercept, Optimizer).

foreign(initModelWithTrain, c, initModelWithTrainI(     +pointer(float_array), +integer, +integer, 
                                                        +pointer(float_array), +integer, 
                                                        +integer, +float32, +float32, 
                                                        +integer, 
                                                        +string)).


%% --Input--
%%              int     numClasses      => 2,
%%              float32 lambda          => 0.0001,
%%              float32 delta           => 1.0,
%%              bool    fitIntercept    => (1)true / (0)false => false,
%%
%% --Output--
%%
%% --Description--
%%              Initializes the linear_svm model with the given data but doesnt train it.
%%
linear_SVM_initModelNoTrain(NumClasses, Lambda, Delta, FitIntercept) :-
        NumClasses >= 0,
        Lambda >= 0.0,
        Delta >= 0.0,
        initModelNoTrainI(NumClasses, Lambda, Delta, FitIntercept).

foreign(initModelNoTrain, c, initModelNoTrainI( +integer, +float32, +float32, 
                                                +integer)).


%% --Input--
%%              mat     data
%%
%% --Output--
%%              vec     labels,
%%              mat     scores
%%
%% --Description--
%%              Classify the given points, returning class scores and predicted class label for each point.
%%
linear_SVM_classify(DataList, DataRows, LabelsList, ScoresList, ZCols) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrows, X)),
        classifyI(X, Xsize, Xrows, Y, Ysize, Z, ZCols, ZRows),
        convert_float_array_to_list(Y, Ysize, LabelsList),
        convert_float_array_to_2d_list(Z, ZCols, ZRows, ScoresList).

foreign(classify, c, classifyI( +pointer(float_array), +integer, +integer, 
                                -pointer(float_array), -integer, 
                                -pointer(float_array), -integer, -integer)).


%% --Input--
%%              vec     point
%%
%% --Output--
%%              int     predicLabel
%%
%% --Description--
%%              Classify the given point.
%%
linear_SVM_classifyPoint(DataList, Prediction) :-
        convert_list_to_float_array(DataList, array(Xsize, X)),
        classifyPointI(X, Xsize, Prediction).

foreign(classifyPoint, c, classifyPointI(       +pointer(float_array), +integer, 
                                                [-integer])).


%% --Input--
%%              mat     testData,
%%              vec     testLabels
%%
%% --Output--
%%              float32 accuracy
%%
%% --Description--
%%              Computes accuracy of the learned model given the feature data and the labels associated with each data point.
%%              Predictions are made using the provided data and are compared with the actual labels.
%%
linear_SVM_computeAccuracy(DataList, DataRows, LabelsList, Accuracy) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        computeAccuracyI(X, Xsize, Xrownum, Y, Ysize, Accuracy).

foreign(computeAccuracy, c, computeAccuracyI(   +pointer(float_array), +integer, +integer, 
                                                +pointer(float_array), +integer, 
                                                [-float32])).


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses      => 2,
%%              string  optimizer       => "lbfgs", "psgd" => lbfgs
%%
%% --Output--
%%              float32 Objective value of the final point
%%
%% --Description--
%%              Train the Linear_svm model with the given training data.
%%
linear_SVM_train(DataList, DataRows, LabelsList, NumClasses, Optimizer, ObjValue) :-
        NumClasses >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        trainI(X, Xsize, Xrownum, Y, Ysize, NumClasses, Optimizer, ObjValue).

foreign(train, c, trainI(       +pointer(float_array), +integer, +integer, 
                                +pointer(float_array), +integer, 
                                +integer, 
                                +string, 
                                [-float32])).


%% Defines the functions that get connected from main.cpp
foreign_resource(linear_SVM, [  initModelWithTrain,
                                initModelNoTrain,
                                classify,
                                classifyPoint,
                                computeAccuracy,
                                train]).

:- load_foreign_resource(linear_SVM).
