#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "linear_regression_glue.h"
#include <mlpack/methods/linear_regression/linear_regression.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::regression;

// Global Variable of the LinearRegression object so it can be accessed from all functions
LinearRegression regressor;
int trainingDimension = 0;

// input:	const arma::mat 	&data, 
//			const arma::rowvec 	&responses, 
//			const double 		lambda, 
//			const bool 			intercept
// output:
void initModel(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
				float *responsesArr, SP_integer responsesArrSize, 
				double lambda, 
				SP_integer bool_intercept)
{
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if labels and weights fit the data
    if (data.n_cols != responsesArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }
	
	// convert the Prolog array to arma::rowvec
	rowvec responsesVector = convertArrayToRowvec(responsesArr, responsesArrSize);


	try
    {
        regressor = LinearRegression(data, responsesVector, lambda, bool_intercept);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
	trainingDimension = data.n_rows;
}

// input:	const arma::mat &data, 
//			const arma::rowvec &responses, 
//			const arma::rowvec &weights, 
//			const double lambda, 
//			const bool intercept
// output:
void initModelWithWeights(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
							float *responsesArr, SP_integer responsesArrSize, 
							float *weightsArr, SP_integer weightsArrSize, 
							double lambda, 
							SP_integer bool_intercept)
{
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if labels and weights fit the data
    if (data.n_cols != responsesArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }
    if (data.n_cols != weightsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of weights!");
        return;
    }

	// convert the Prolog array to arma::rowvec
	rowvec responsesVector = convertArrayToRowvec(responsesArr, responsesArrSize);
	rowvec weightsVector = convertArrayToRowvec(weightsArr, weightsArrSize);


	try
    {
        regressor = LinearRegression(data, responsesVector, weightsVector, lambda, bool_intercept);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
	trainingDimension = data.n_rows;
}

// input: 	const arma::mat &points, 
//			const arma::rowvec &responses
// output: double L2SquardError
double computeError(float *pointsMatArr, SP_integer pointsMatSize, SP_integer pointsMatRowNum, 
					float *responsesArr, SP_integer responsesArrSize)
{
	// convert the Prolog array to arma::mat
	mat points = convertArrayToMat(pointsMatArr, pointsMatSize, pointsMatRowNum);
	// check if labels fit the data
    if (points.n_cols != responsesArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }

	// convert the Prolog array to arma::rowvec
	rowvec responsesVector = convertArrayToRowvec(responsesArr, responsesArrSize);


	// run the model function
	try
    {
        return regressor.ComputeError(points, responsesVector);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return 0.0;
    }
}

// input:
// output: const arma::vec & parameters b vector
void parameters(float **bArr, SP_integer *bArrSize)
{
	// create the ReturnVector
	vec bReturnVector = regressor.Parameters();
	
	// return the Vector
	returnVectorInformation(bReturnVector, bArr, bArrSize);
}

// input: arma::vec & new b vector
// output:
void modifyParameters(float *newBArr, SP_integer newBArrSize)
{
	// convert the Prolog array to arma::rowvec
	vec newBVector = convertArrayToColvec(newBArr, newBArrSize);


	// get parameter b vector ref
	vec& bVector = regressor.Parameters();

	// set new b vector
	bVector = newBVector;
}

// input: 	const arma::mat &points, 
//			arma::rowvec &predictions
// output:
void predict(float *pointsMatArr, SP_integer pointsMatSize, SP_integer pointsMatRowNum, 
				float **predicArr, SP_integer *predicArrSize)
{
	// convert the Prolog array to arma::mat
	mat points = convertArrayToMat(pointsMatArr, pointsMatSize, pointsMatRowNum);
	if (points.n_rows != trainingDimension)
	{
		raisePrologSystemExeption("The given Points have a diffrent Dimension than the trained Model!");
        return;
	}
	

	// create the ReturnVector
	rowvec predicReturnVector;
	

	// run the predict function on the model
	try
    {
        regressor.Predict(points, predicReturnVector);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
	

	// return the Vector
	returnVectorInformation(predicReturnVector, predicArr, predicArrSize);
}

// input: 	const arma::mat &data, 
//			const arma::rowvec &responses, 
//			const bool intercept
// output:
double train(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
				float *responsesArr, SP_integer responsesArrSize, 
				SP_integer bool_intercept)
{
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if labels fit the data
    if (data.n_cols != responsesArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }

	// convert the Prolog array to arma::rowvec
	rowvec responsesVector = convertArrayToRowvec(responsesArr, responsesArrSize);


	// run the model train function
	try
    {
        return regressor.Train(data, responsesVector, (bool_intercept == 1));
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return 0.0;
    }
	trainingDimension = data.n_rows;
}

// input: 	const arma::mat 	&data, 
//			const arma::rowvec 	&responses, 
//			const arma::rowvec 	&weights, 
//			const bool 			intercept
// output:
double trainWithWeights(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
						float *responsesArr, SP_integer responsesArrSize, 	
						float *weightsArr, SP_integer weightsArrSize, 
						SP_integer bool_intercept)
{
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);	
	// check if labels and weights fit the data
    if (data.n_cols != responsesArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }
    if (data.n_cols != weightsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of weights!");
        return 0.0;
    }

	// convert the Prolog array to arma::rowvec
	rowvec responsesVector = convertArrayToRowvec(responsesArr, responsesArrSize);
	rowvec weightsVector = convertArrayToRowvec(weightsArr, weightsArrSize);

	// run the model train function
	try
    {
        return regressor.Train(data, responsesVector, weightsVector, (bool_intercept == 1));
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return 0.0;
    }
	trainingDimension = data.n_rows;
}
