
:- module(linear_regression, [  linear_regression_initModel/5, 
                                linear_regression_initModelWithWeights/6, 
                                linear_regression_computeError/4, 
                                linear_regression_parameters/1, 
                                linear_regression_modifyParameters/1, 
                                linear_regression_predict/3, 
                                linear_regression_train/5, 
                                linear_regression_trainWithWeights/6]).

:- load_files(library(str_decl),
                [when(compile_time), if(changed)]).

%% needed for using the array type and for reading from csv
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function and what there inputs and output arguments are


%% --Input--
%%              mat     data,
%%              vec     responses,
%%              float32 lambda          => 0.0,
%%              bool    intercept       => (1)true / (0)false => true,
%%
%% --Output--
%%
%% --Description--
%%              Initializes the linear_regression model and trains it but doesnt include weights.
%%
linear_regression_initModel(DataList, DataRows, ResponsesList, Lambda, Intercept) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(ResponsesList, array(Ysize, Y)),
        initModelI(X, Xsize, Xrownum, Y, Ysize, Lambda, Intercept).

foreign(initModel,  c, initModelI(      +pointer(float_array), +integer, +integer, 
                                        +pointer(float_array), +integer, 
                                        +float32, 
                                        +integer)).


%% --Input--
%%              mat     data,
%%              vec     responses,
%%              vec     weights,
%%              float32 lambda          => 0.0,
%%              bool    intercept       => (1)true / (0)false => true,
%%
%% --Output--
%%
%% --Description--
%%              Initializes the linear_regression model, trains it and adds weights to it.
%%
linear_regression_initModelWithWeights(DataList, DataRows, ResponsesList, WeightsList, Lambda, Intercept) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(ResponsesList, array(Ysize, Y)),
        convert_list_to_float_array(WeightsList, array(Zsize, Z)),
        initModelWithWeightsI(X, Xsize, Xrownum, Y, Ysize, Z, Zsize, Lambda, Intercept).

foreign(initModelWithWeights,  c, initModelWithWeightsI(        +pointer(float_array), +integer, +integer, 
                                                                +pointer(float_array), +integer, 
                                                                +pointer(float_array), +integer, 
                                                                +float32, 
                                                                +integer)).


%% --Input--
%%              mat     points,
%%              vec     responses
%%
%% --Output--
%%              float32 error
%%
%% --Description--
%%              Calculate the L2 squared error on the given predictors and responses using this linear regression model.
%%
linear_regression_computeError(DataList, DataRows, ResponsesList, Error) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(ResponsesList, array(Ysize, Y)),
        computeErrorI(X, Xsize, Xrownum, Y, Ysize, Error).

foreign(computeError,  c, computeErrorI(+pointer(float_array), +integer, +integer, 
                                        +pointer(float_array), +integer, 
                                        [-float32])).


%% --Input--
%%
%% --Output--
%%              vec     parameters
%%
%% --Description--
%%              Get the parameters (the b vector).
%%
linear_regression_parameters(ParametersList) :-
        parametersI(Y, Ysize),
        convert_float_array_to_list(Y, Ysize, ParametersList).

foreign(parameters, c, parametersI(-pointer(float_array), -integer)).


%% --Input--
%%              vec     parameters
%%
%% --Output--
%%
%% --Description--
%%              Modify the parameters (the b vector).
%%
linear_regression_modifyParameters(ParameterList) :-
        convert_list_to_float_array(ParameterList, array(Ysize, Y)),
        modifyParametersI(Y, Ysize).

foreign(modifyParameters,  c, modifyParametersI(+pointer(float_array), +integer)).


%% --Input--
%%              mat     points
%%
%% --Output--
%%              vec     responses
%%
%% --Description--
%%              Calculate y_i for each data point in points.
%%
linear_regression_predict(PointsList, PointsRows, PredicList) :-
        convert_list_to_float_array(PointsList, PointsRows, array(Xsize, Xrownum, X)),
        predictI(X, Xsize, Xrownum, Y, Ysize),
        convert_float_array_to_list(Y, Ysize, PredicList).

foreign(predict,  c, predictI(  +pointer(float_array), +integer, +integer, 
                                -pointer(float_array), -integer)).


%% --Input--
%%              mat     data,
%%              vec     responses,
%%              bool    intercept       => (1)true / (0)false => true,
%%
%% --Output--
%%              float32 least quares error after training
%%
%% --Description--
%%              Train the linear_regression model on the given data.
%%              Careful! 
%%              This will completely ignore and overwrite the existing model. This particular implementation does not have an incremental training algorithm.
%%
linear_regression_train(DataList, DataRows, ResponsesList, Intercept, Error) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(ResponsesList, array(Ysize, Y)),
        trainI(X, Xsize, Xrownum, Y, Ysize, Intercept, Error).

foreign(train,  c, trainI(      +pointer(float_array), +integer, +integer, 
                                +pointer(float_array), +integer, 
                                +integer, 
                                [-float32])).


%% --Input--
%%              mat     data,
%%              vec     responses,
%%              vec     weights,
%%              bool    intercept       => (1)true / (0)false => true,
%%
%% --Output--
%%              float32 least quares error after training
%%
%% --Description--
%%              Train the linear_regression model on the given data and weights.
%%              Careful! 
%%              This will completely ignore and overwrite the existing model. This particular implementation does not have an incremental training algorithm.
%%
linear_regression_trainWithWeights(DataList, DataRows, ResponsesList, WeightsList, Intercept, Error) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(ResponsesList, array(Ysize, Y)),
        convert_list_to_float_array(WeightsList, array(Zsize, Z)),
        trainWithWeightsI(X, Xsize, Xrownum, Y, Ysize, Z, Zsize, Intercept, Error).

foreign(trainWithWeights,  c, trainWithWeightsI( +pointer(float_array), +integer, +integer, 
                                                +pointer(float_array), +integer, 
                                                +pointer(float_array), +integer, 
                                                +integer, 
                                                [-float32])).


%% Defines what functions should be connected from main.cpp
foreign_resource(linear_regression, [   initModel, 
                                        initModelWithWeights, 
                                        computeError, 
                                        parameters, 
                                        modifyParameters, 
                                        predict, 
                                        train, 
                                        trainWithWeights]).

:- load_foreign_resource(linear_regression).
