#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "lmnn_glue.h"
#include <mlpack/methods/lmnn/lmnn.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace lmnn;
using namespace mlpack::metric;


// input:   const arma::mat & 	            dataset,
//          const arma::Row< size_t > & 	labels,
//          const size_t 	                k,
//          const MetricType 	            metric = MetricType()
//          arma::mat & 	                outputMatrix <-,
// output: 
// description: 
//          Is a single predicate that initiates the lmnn model with all the given params and then performs Large Margin Nearest Neighbors metric learning on the reference data.
//
void lmnn(char const *optimizer, 
            float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum,
            float *labelsArr, SP_integer labelsArrSize,
            SP_integer k,
            double regularization, double stepSize, SP_integer passes, SP_integer maxIterations, double tolerance, 
            SP_integer center, SP_integer shuffle, 
            SP_integer batchSize, SP_integer range, SP_integer rank,
            float **outputMatArr, SP_integer *outputMatColNum, SP_integer *outputMatRowNum)
{
    // convert the Prolog arrays to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    // check if labels fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }
    // convert the Prolog arrays to arma::rowvec
    Row< size_t > labelsVector = convertArrayToVec(labelsArr, labelsArrSize);


    // Carry out mean-centering on the dataset, if necessary.
    if ((center == 1))
    {
        for (size_t i = 0; i < data.n_rows; ++i)
        {
            data.row(i) -= arma::mean(data.row(i));
        }
    }

    // Now, normalize the labels.
    arma::Col<size_t> mappings;
    arma::Row<size_t> labels;
    data::NormalizeLabels(labelsVector, labels, mappings);
    
    // create the ReturnMat
    mat outputReturnMat;

    if (rank > 0)
    {
        outputReturnMat = arma::randu(rank, data.n_rows);
    }
    else
    {
        outputReturnMat.eye();
    }

    
    try
    {
        if (strcmp(optimizer, "amsgrad") == 0)
        {
            LMNN<LMetric<2>> lmnn(data, labelsVector, k);
            lmnn.Regularization() = regularization;
            lmnn.Range() = range;
            lmnn.Optimizer().StepSize() = stepSize;
            lmnn.Optimizer().MaxIterations() = passes * data.n_cols;
            lmnn.Optimizer().Tolerance() = tolerance;
            lmnn.Optimizer().Shuffle() = (shuffle == 1);
            lmnn.Optimizer().BatchSize() = batchSize;

            lmnn.LearnDistance(outputReturnMat);
        }
        else if (strcmp(optimizer, "bbsgd") == 0)
        {
            LMNN<LMetric<2>, ens::BBS_BB> lmnn(data, labelsVector, k);
            lmnn.Regularization() = regularization;
            lmnn.Range() = range;
            lmnn.Optimizer().StepSize() = stepSize;
            lmnn.Optimizer().MaxIterations() = passes * data.n_cols;
            lmnn.Optimizer().Tolerance() = tolerance;
            lmnn.Optimizer().Shuffle() = (shuffle == 1);
            lmnn.Optimizer().BatchSize() = batchSize;

            lmnn.LearnDistance(outputReturnMat);
        }
        else if (strcmp(optimizer, "sgd") == 0)
        {
            // copied advise from the lmnn.main file
            // Using SGD is not recommended as the learning matrix can
            // diverge to inf causing serious memory problems.
            LMNN<LMetric<2>, ens::StandardSGD> lmnn(data, labelsVector, k);
            lmnn.Regularization() = regularization;
            lmnn.Range() = range;
            lmnn.Optimizer().StepSize() = stepSize;
            lmnn.Optimizer().MaxIterations() = passes * data.n_cols;
            lmnn.Optimizer().Tolerance() = tolerance;
            lmnn.Optimizer().Shuffle() = (shuffle == 1);
            lmnn.Optimizer().BatchSize() = batchSize;

            lmnn.LearnDistance(outputReturnMat);
        }
        else if (strcmp(optimizer, "lbfgs") == 0)
        {
            LMNN<LMetric<2>, ens::L_BFGS> lmnn(data, labelsVector, k);
            lmnn.Regularization() = regularization;
            lmnn.Range() = range;
            lmnn.Optimizer().MaxIterations() = maxIterations;
            lmnn.Optimizer().MinGradientNorm() = tolerance;

            lmnn.LearnDistance(outputReturnMat);
        }
        else
        {
            raisePrologDomainExeption(optimizer, 1, "The given Optimizer is unkown!", "lmnn");
        }
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    
    
    // return the Matrix
    returnMatrixInformation(outputReturnMat, outputMatArr, outputMatColNum, outputMatRowNum);
}
