#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "logistic_regression_glue.h"
#include <mlpack/methods/logistic_regression/logistic_regression.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::regression;

// Global Variable of the LogisticRegression object so it can be accessed from all functions
LogisticRegression logisticRegression;

int trainedDimension = 0;


// input: 	const MatType & 				predictors,
//			const arma::Row< size_t > & 	responses,
//			const arma::rowvec & 			initialPoint,
//			const double 					lambda = 0
// output: 
// description: 
//			Initialize the logistic_regression model without specifing a optimizer.
//
void initModelNoOptimizer(float *predictorsMatArr, SP_integer predictorsMatSize, SP_integer predictorsMatRowNum, 
							float *responsesArr, SP_integer responsesArrSize, 
							double lambda)
{
	// convert the Prolog array to arma::mat
	mat predictors = convertArrayToMat(predictorsMatArr, predictorsMatSize, predictorsMatRowNum);
	// check if labels fit the data
    if (predictors.n_cols != responsesArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }

	// convert the Prolog array to arma::rowvec
	arma::Row<size_t> responsesVector = convertArrayToVec(responsesArr, responsesArrSize);

	
	try
	{
		logisticRegression = LogisticRegression(predictors, responsesVector, lambda);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}
	trainedDimension = predictors.n_rows;
}


// input: 	const MatType & 				predictors,
//			const arma::Row< size_t > & 	responses,
//			OptimizerType & 				optimizer,
//			const double 					lambda 
// output: 
// description: 
//			Initialize the logistic_regression model and specify the optimizer.
//
void initModelWithOptimizer(float *predictorsMatArr, SP_integer predictorsMatSize, SP_integer predictorsMatRowNum, 
							float *responsesArr, SP_integer responsesArrSize, 
							char const *optimizer, 
							double lambda, SP_integer maxIterations, double tolerance, double stepSize, SP_integer batchSize)
{
	// convert the Prolog array to arma::mat
	mat predictors = convertArrayToMat(predictorsMatArr, predictorsMatSize, predictorsMatRowNum);
	if (predictors.n_cols != responsesArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }

	// convert the Prolog array to arma::rowvec
	arma::Row<size_t> responsesVector = convertArrayToVec(responsesArr, responsesArrSize);

	try
	{
		logisticRegression = LogisticRegression<>(predictors.n_cols, lambda);
	
		if (strcmp(optimizer, "lbfgs") == 0)
		{
			ens::L_BFGS lbfgsOpt;
      		lbfgsOpt.MaxIterations() = maxIterations;
      		lbfgsOpt.MinGradientNorm() = tolerance;
      		Log::Info << "Training model with L-BFGS optimizer." << endl;

      		// This will train the model.
      		logisticRegression.Train(predictors, responsesVector, lbfgsOpt);
		}
		else if (strcmp(optimizer, "psgd") == 0)
		{
			ens::SGD<> sgdOpt;
      		sgdOpt.MaxIterations() = maxIterations;
      		sgdOpt.Tolerance() = tolerance;
      		sgdOpt.StepSize() = stepSize;
      		sgdOpt.BatchSize() = batchSize;
      		Log::Info << "Training model with SGD optimizer." << endl;

      		// This will train the model.
      		logisticRegression.Train(predictors,responsesVector, sgdOpt);
		}
		else
		{
			raisePrologDomainExeption(optimizer, 4, "The given Optimizer is unkown!", "initModelWithOptimizer");
		}
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}
	trainedDimension = predictors.n_rows;
}


// input: 	const VecType & 	point,
//			const double 		decisionBoundary = 0.5
//
// output: 	size_t				predicted point label
// description: 
//			Classify the given point.
//
SP_integer classifyPoint(float *pointArr, SP_integer pointArrSize, 
							double decisionBoundary)
{
	if (pointArrSize != trainedDimension)
	{
		raisePrologSystemExeption("The given Datapoint has a different Dimension than the trained Model!");
        return 0;
	}
	// convert the Prolog array to arma::rowvec
	rowvec pointVector = convertArrayToRowvec(pointArr, pointArrSize);
	try
	{
		return logisticRegression.Classify(pointVector, decisionBoundary);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return 0;
	}
}


// input: 	const MatType & 		dataset,
//			arma::Row< size_t > & 	labels <-,
//			arma::mat &				probabilities <-,
//			const double 			decisionBoundary = 0.5
// output: 
// description: 
//			Classify the given points, returning the predicted labels and probabilities for each point.
//
void classifyMatrix(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
					float **labelsArr, SP_integer *labelsArrSize, 
					float **probsMatArr, SP_integer *probsMatColNum, SP_integer *probsMatRowNum, 
					double decisionBoundary)
{
	
	// convert the Prolog arrays to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	if (data.n_rows != trainedDimension)
	{
		raisePrologSystemExeption("The given Datapoints have a different Dimension than the trained Model!");
        return;
	}
	
	// get the ReturnVector
	Row<size_t> labelsReturnVector;
	// create the ReturnMat
	mat probsReturnMat;
	

	try
	{
		logisticRegression.Classify(data, labelsReturnVector, decisionBoundary);
		logisticRegression.Classify(data, probsReturnMat);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}
	
	
	// return the Vector
	returnVectorInformation(labelsReturnVector, labelsArr, labelsArrSize);
	// return the Matrix
	returnMatrixInformation(probsReturnMat, probsMatArr, probsMatColNum, probsMatRowNum);
}


// input: 	const MatType & 				predictors,
//			const arma::Row< size_t > & 	responses,
//			const double 					decisionBoundary = 0.5
//
// output: 	double							accuracy
// description: 
//			Compute the accuracy of the model on the given predictors and responses, using the given decision boundary.
//
double computeAccuracy(float *predictorsMatArr, SP_integer predictorsMatSize, SP_integer predictorsMatRowNum, 
						float *responsesArr, SP_integer responsesArrSize, 
						double decisionBoundary)
{
	// convert the Prolog array to arma::mat
	mat predictors = convertArrayToMat(predictorsMatArr, predictorsMatSize, predictorsMatRowNum);
	if (predictors.n_cols != responsesArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }
	if (predictors.n_rows != trainedDimension)
	{
		raisePrologSystemExeption("The given Datapoints have a different Dimension than the trained Model!");
        return 0.0;
	}

	// convert the Prolog array to arma::rowvec
	arma::Row<size_t> responsesVector = convertArrayToVec(responsesArr, responsesArrSize);

	try
	{
		return logisticRegression.ComputeAccuracy(predictors, responsesVector, decisionBoundary);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return 0.0;
	}
}


// input: 	const MatType & 				predictors,
//			const arma::Row< size_t > & 	responses
//
// output: 	double							error
// description: 
//			Compute the error of the model.
//
double computeError(float *predictorsMatArr, SP_integer predictorsMatSize, SP_integer predictorsMatRowNum, 
					float *responsesArr, SP_integer responsesArrSize)
{
	// convert the Prolog arrays to arma::mat
	mat predictors = convertArrayToMat(predictorsMatArr, predictorsMatSize, predictorsMatRowNum);
	if (predictors.n_cols != responsesArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }
	if (predictors.n_rows != trainedDimension)
	{
		raisePrologSystemExeption("The given Datapoints have a different Dimension than the trained Model!");
        return 0.0;
	}

	// convert the Prolog array to arma::rowvec
	arma::Row<size_t> responsesVector = convertArrayToVec(responsesArr, responsesArrSize);

	double error = 0.0;
	try
	{
		error = logisticRegression.ComputeError(predictors, responsesVector);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return 0.0;
	}
	if (error != error)
	{
		raisePrologSystemExeption("Error contains nan Value!");
		return 0.0;
	}
	
	return error;
}