#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "lsh_glue.h"
#include <mlpack/methods/lsh/lsh_search.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::neighbor;

// Global Variable of the LSHSearch object so it can be accessed from all functions
LSHSearch lshSearch;

// TODO: adding support for arma::cube


// input:   MatType 	    referenceSet,
//          const size_t 	numProj,
//          const size_t 	numTables,
//          const double 	hashWidth = 0.0,
//          const size_t 	secondHashSize = 99901,
//          const size_t 	bucketSize = 500
// output: 
// description: 
//          Initiatzes the model and trains it.
//
void initAndTrainModel(float *referenceMatArr, SP_integer referenceMatSize, SP_integer referenceMatRowNum,
                SP_integer numProj, SP_integer numTables, double hashWidth, SP_integer secondHashSize, SP_integer bucketSize)
{
    // convert the Prolog array to arma::mat
    mat reference = convertArrayToMat(referenceMatArr, referenceMatSize, referenceMatRowNum);

    try
    {
        lshSearch = LSHSearch(reference, numProj, numTables, hashWidth, secondHashSize, bucketSize);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
}


// input:   const arma::Mat< size_t > & 	foundNeighbors,
//          const arma::Mat< size_t > & 	realNeighbors 
// output:  double  recall percentage [0, 1]
// description: 
//          Compute the recall (% of neighbors found) given the neighbors returned by searchWithQuery/12 or searchNoQuery/9 and a "ground truth" set of neighbors.
//
double computeRecall(float *foundNeighborsMatArr, SP_integer foundNeighborsMatSize, SP_integer foundNeighborsMatRowNum, 
                        float *realNeighborsMatArr, SP_integer realNeighborsMatSize, SP_integer realNeighborsMatRowNum)
{
    // convert the Prolog array to arma::mat
    Mat< size_t > foundNeighbors = conv_to<Mat< size_t >>::from(convertArrayToMat(foundNeighborsMatArr, foundNeighborsMatSize, foundNeighborsMatRowNum));
    Mat< size_t > realNeighbors = conv_to<Mat< size_t >>::from(convertArrayToMat(realNeighborsMatArr, realNeighborsMatSize, realNeighborsMatRowNum));

    try
    {
        return lshSearch.ComputeRecall(foundNeighbors, realNeighbors);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return 0.0;
    }
}


// input:   const MatType & 	    querySet,
//          const size_t 	        k,
//          arma::Mat< size_t > & 	resultingNeighbors <-,
//          arma::mat & 	        distances <-,
//          const size_t 	        numTablesToSearch = 0,
//          const size_t 	        T = 0 
// output: 
// description: 
//          Compute the nearest neighbors of the points in the given query set and store the output in the given matrices.
//
void searchWithQuery(float *querySetMatArr, SP_integer querySetMatSize, SP_integer querySetMatRowNum, 
                            SP_integer k, 
                            float **resultingNeighborsMatArr, SP_integer *resultingNeighborsMatColNum, SP_integer *resultingNeighborsMatRowNum,
                            float **distancesMatArr, SP_integer *distancesMatColNum, SP_integer *distancesMatRowNum,
                            SP_integer numTablesToSearch, SP_integer T)
{
    // convert the Prolog array to arma::mat
    mat querySet = convertArrayToMat(querySetMatArr, querySetMatSize, querySetMatRowNum);
    if (querySet.n_cols < k)
    {
        raisePrologSystemExeption("K should not be greater than the amount of datapoints!");
        return;
    }
    

    // create the ReturnMat
    Mat< size_t > resultingNeighborsReturnMat;
    mat distancesReturnMat;


    try
    {
        lshSearch.Search(querySet, k, resultingNeighborsReturnMat, distancesReturnMat, numTablesToSearch, T);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }

    // float32 cant handle values over 1.0e+38
    distancesReturnMat = clamp(distancesReturnMat, -1.0e+38, 1.0e+38);
    

    // return the Matrices
    returnMatrixInformation(resultingNeighborsReturnMat, resultingNeighborsMatArr, resultingNeighborsMatColNum, resultingNeighborsMatRowNum);
    returnMatrixInformation(distancesReturnMat, distancesMatArr, distancesMatColNum, distancesMatRowNum);
}


// input:   const size_t 	        k,
//          arma::Mat< size_t > & 	resultingNeighbors <-,
//          arma::mat & 	        distances <-,
//          const size_t 	        numTablesToSearch = 0,
//          size_t 	                T = 0 
// output: 
// description: 
//          Compute the nearest neighbors and store the output in the given matrices.
//
void searchNoQuery(SP_integer k, 
                    float **resultingNeighborsMatArr, SP_integer *resultingNeighborsMatColNum, SP_integer *resultingNeighborsMatRowNum,
                    float **distancesMatArr, SP_integer *distancesMatColNum, SP_integer *distancesMatRowNum,
                    SP_integer numTablesToSearch, SP_integer T)
{
    if (lshSearch.ReferenceSet().n_cols < k)
    {
        raisePrologSystemExeption("K should not be greater than the amount of datapoints!");
        return;
    }
    // create the ReturnMat
    Mat< size_t > resultingNeighborsReturnMat;
    mat distancesReturnMat;

    try
    {
        lshSearch.Search(k, resultingNeighborsReturnMat, distancesReturnMat, numTablesToSearch, T);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }

    // float32 cant handle values over 1.0e+38
    distancesReturnMat = clamp(distancesReturnMat, -1.0e+38, 1.0e+38);

    // return the Matrices
    returnMatrixInformation(resultingNeighborsReturnMat, resultingNeighborsMatArr, resultingNeighborsMatColNum, resultingNeighborsMatRowNum);
    returnMatrixInformation(distancesReturnMat, distancesMatArr, distancesMatColNum, distancesMatRowNum);
}
