
:- module(lsh, [lsh_initAndTrainModel/7,
                lsh_computeRecall/5,
                lsh_searchWithQuery/9,
                lsh_searchNoQuery/7]).

%% requirements of library(struct)
:- load_files(library(str_decl),
        [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function


%% --Input--
%%              mat     referenceSet,
%%              int     numProj         => 10-50,
%%              int     numTables       => 10-20,
%%              float32 hashWidth       => 0.0,
%%              int     secondHashSize  => 99901,
%%              int     bucketSize      => 500
%%
%% --Output--
%%
%% --Description--
%%              Initiatzes the model and trains it.
%%
lsh_initAndTrainModel(ReferenceList, ReferenceRows, NumProj, NumTables, HashWidth, SecondHashSize, BucketSize) :-
        NumProj >= 0,
        NumTables >= 0,
        HashWidth >= 0.0,
        SecondHashSize > 0,
        BucketSize > 0,
        convert_list_to_float_array(ReferenceList, ReferenceRows, array(Xsize, Xrownum, X)),
        initAndTrainModelI(X, Xsize, Xrownum, NumProj, NumTables, HashWidth, SecondHashSize, BucketSize).

foreign(initAndTrainModel, c, initAndTrainModelI(+pointer(float_array), +integer, +integer,
                                 +integer, +integer, +float32, +integer, +integer)).


%% --Input--
%%              mat     foundNeighbors,
%%              mat     realNeighbors
%%
%% --Output--
%%              float32 recall percentage       => values in between [0,1]
%%
%% --Description--
%%              Compute the recall (% of neighbors found) given the neighbors returned by searchWithQuery/12 or searchNoQuery/9 and a "ground truth" set of neighbors.
%%
lsh_computeRecall(FoundNeighborsList, FoundNeighborsRows, RealNeighborsList, RealNeighborsRows, Percentage) :-
        convert_list_to_float_array(FoundNeighborsList, FoundNeighborsRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(RealNeighborsList, RealNeighborsRows, array(Ysize, Yrownum, Y)),
        computeRecallI(X, Xsize, Xrownum, Y, Ysize, Yrownum, Percentage).

foreign(computeRecall, c, computeRecallI(+pointer(float_array), +integer, +integer,
                                         +pointer(float_array), +integer, +integer,
                                         [-float32])).


%% --Input--
%%              mat     querySet,
%%              int     k                       => 0,
%%              int     numTablesToSearch       => 0,
%%              int     T                       => 0
%%
%% --Output--
%%              mat     resultingNeighbors,
%%              mat     distances
%%
%% --Description--
%%              Compute the nearest neighbors of the points in the given query set and store the output in the given matrices.
%%              The matrices will be set to the size of n columns by k rows, where n is the number of points in the query dataset and k is the number of neighbors being searched for.
%%
lsh_searchWithQuery(QueryList, QueryRows, K, ResultingNeighborsList, YCols, DistancesList, ZCols, NumTablesToSearch, T) :-
        K > 0,
        NumTablesToSearch >= 0,
        T >= 0,
        convert_list_to_float_array(QueryList, QueryRows, array(Xsize, Xrows, X)),
        searchWithQueryI(X, Xsize, Xrows, K, Y, YCols, YRows, Z, ZCols, ZRows, NumTablesToSearch, T),
        convert_float_array_to_2d_list(Y, YCols, YRows, ResultingNeighborsList),
        convert_float_array_to_2d_list(Z, ZCols, ZRows, DistancesList).

foreign(searchWithQuery, c, searchWithQueryI(   +pointer(float_array), +integer, +integer, 
                                                +integer, 
                                                -pointer(float_array), -integer, -integer, 
                                                -pointer(float_array), -integer, -integer, 
                                                +integer, +integer)).


%% --Input--
%%              int     k                       => 0,
%%              int     numTablesToSearch       => 0,
%%              int     T                       => 0
%%
%% --Output--
%%              mat     resultingNeighbors,
%%              mat     distances
%%
%% --Description--
%%              Compute the nearest neighbors and store the output in the given matrices.
%%              The matrices will be set to the size of n columns by k rows, where n is the number of points in the query dataset and k is the number of neighbors being searched for.
%%
lsh_searchNoQuery(K, ResultingNeighborsList, YCols, DistancesList, ZCols, NumTablesToSearch, T) :-
        K > 0,
        NumTablesToSearch >= 0,
        T >= 0,
        searchNoQueryI(K, Y, YCols, YRows, Z, ZCols, ZRows, NumTablesToSearch, T),
        convert_float_array_to_2d_list(Y, YCols, YRows, ResultingNeighborsList),
        convert_float_array_to_2d_list(Z, ZCols, ZRows, DistancesList).

foreign(searchNoQuery, c, searchNoQueryI(       +integer, 
                                                -pointer(float_array), -integer, -integer, 
                                                -pointer(float_array), -integer, -integer, 
                                                +integer, +integer)).



%% Defines the functions that get connected from main.cpp
foreign_resource(lsh, [ initAndTrainModel,
                        computeRecall,
                        searchWithQuery,
                        searchNoQuery]).

:- load_foreign_resource(lsh).
