#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "naive_bayes_classifier_glue.h"
#include <mlpack/methods/naive_bayes/naive_bayes_classifier.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::naive_bayes;

// Global Variable of the NaiveBayesClassifier object so it can be accessed from all functions
NaiveBayesClassifier<mat> naiveBayesClassifier;
bool isModelTrained = false;
int trainedDimensionality = 0;


// input: 	const MatType & 				data,
//			const arma::Row< size_t > & 	labels,
//			const size_t 					numClasses,
//			const bool 						incrementalVariance = false,
//			const double 					epsilon = 1e-10 
// output: 
// description: 
//			Initializes the classifier as per the input and then trains it by calculating the sample mean and variances.
//
void initModelWithTrain(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
						float *labelsArr, SP_integer labelsArrSize, 
						SP_integer numClasses, 
						SP_integer incrementalVariance, 
						double epsilon)
{
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if labels fit the data
	if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }

	// convert the Prolog array to arma::rowvec
	Row<size_t> labelsVector = convertArrayToVec(labelsArr, labelsArrSize);

	try
	{
		naiveBayesClassifier = NaiveBayesClassifier(data, labelsVector, numClasses, (incrementalVariance == 1), epsilon);
	}
    catch(const std::out_of_range& e)
    {
        raisePrologSystemExeption("The given Labels dont fit the format [0,Numclasses-1]!");
        return;
    }
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}
	isModelTrained = true;
	trainedDimensionality = data.n_rows;
}


// input: 	const size_t 	dimensionality = 0,
//			const size_t 	numClasses = 0,
//			const double 	epsilon = 1e-10 
// output: 
// description: 
//			Initialize the Naive Bayes classifier without performing training.
//
void initModelNoTrain(SP_integer dimensionality, SP_integer numClasses, double epsilon)
{
	try
	{
		naiveBayesClassifier = NaiveBayesClassifier(dimensionality, numClasses, epsilon);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}
	isModelTrained = false;
	trainedDimensionality = dimensionality;
}

/*
// TODO: 	creates an error
// input: 	const VecType & 		point,
//			size_t & 				prediction <-,
//			ProbabilitiesVecType & 	probabilities <-
// output: 
// description: 
void classifyPoint(float *pointArr, SP_integer pointArrSize, SP_integer *prediction, float **probsArr, SP_integer *probsArrSize)
{
	// convert the Prolog arrays to arma::rowvec
	vec pointVector = convertArrayToRowvec(pointArr, pointArrSize);

	// get the ReturnVector
	vec probsReturnVector;

	size_t predictionReturn;

	naiveBayesClassifier.Classify(pointVector, predictionReturn, probsReturnVector);

	*prediction = predictionReturn;

	// return the Vector lenght
	*probsArrSize = probsReturnVector.n_elem;

	// return the Vector as Array
	*probsArr = convertToArray(probsReturnVector);
}
*/


// input: 	const MatType & 		data,
//			arma::Row< size_t > & 	predictions <-,
//			ProbabilitiesMatType & 	probabilities <-
// output: 
// description: 
//			Classify the given points using the trained NaiveBayesClassifier model and also return estimates of the probabilities for each class in the given matrix.
//
void classifyMatrix(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
					float **predicArr, SP_integer *predicArrSize, 
					float **probsMatArr, SP_integer *probsMatColNum, SP_integer *probsMatRowNum)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not trained!");
		return;
	}
	// convert the Prolog arrays to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if data Dim fits trained Dim
	if (data.n_rows != trainedDimensionality)
	{
		raisePrologSystemExeption("The Datapoints Dimensionality doesnt fit the trained Dimensionality!");
		return;
	}
	
	
	// get the ReturnVector
	Row<size_t> predicReturnVector;
	// get the ReturnMat
	mat probsReturnMat;


	try
	{
		naiveBayesClassifier.Classify(data, predicReturnVector, probsReturnMat);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}


	// return the Vector
	returnVectorInformation(predicReturnVector, predicArr, predicArrSize);
	// return the Matrix
	returnMatrixInformation(probsReturnMat, probsMatArr, probsMatColNum, probsMatRowNum);
}


// input: 
// output: 	mat	sample means
// description: 
//			Get the sample means for each class.
//
void means(float **meansMatArr, SP_integer *meansMatColNum, SP_integer *meansMatRowNum)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not trained!");
		return;
	}
	// get the ReturnMat
	mat meansReturnMat = naiveBayesClassifier.Means();


	// return the Matrix
	returnMatrixInformation(meansReturnMat, meansMatArr, meansMatColNum, meansMatRowNum);
}


// input: 
// output: 	mat	prior probabilities
// description: 
//			Get the sample probabilities for each class.
//
void probabilities(float **probsMatArr, SP_integer *probsMatColNum, SP_integer *probsMatRowNum)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not trained!");
		return;
	}
	// get the ReturnMat
	mat probsReturnMat = naiveBayesClassifier.Probabilities();


	// return the Matrix
	returnMatrixInformation(probsReturnMat, probsMatArr, probsMatColNum, probsMatRowNum);
}


// input: 	const MatType & 				data,
//			const arma::Row< size_t > & 	labels,
//			const size_t 					numClasses,
//			const bool 						incremental = true
// output: 
// description: 
//			Train the Naive Bayes classifier on the given dataset.
//
void trainMatrix(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
					float *labelsArr, SP_integer labelsArrSize, 
					SP_integer numClasses, 
					SP_integer incrementalVariance)
{
	// convert the Prolog array to arma::mat
	mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
	// check if labels fit the data
	if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }
	// check if data Dim fits trained Dim
	if (data.n_rows != trainedDimensionality)
	{
		raisePrologSystemExeption("The Datapoints Dimensionality doesnt fit the trained Dimensionality!");
		return;
	}

	// convert the Prolog array to arma::rowvec
	Row<size_t> labelsVector = convertArrayToVec(labelsArr, labelsArrSize);


	try
	{
		naiveBayesClassifier.Train(data, labelsVector, numClasses, (incrementalVariance == 1));
	}
    catch(const std::out_of_range& e)
    {
        raisePrologSystemExeption("The given Labels dont fit the format [0,Numclasses-1]!");
        return;
    }
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}

	isModelTrained = true;
}


// input: 	const VecType & 	point,
//			const size_t 		label
// output: 	
// description: 
//			Train the Naive Bayes classifier on the given point.
//
void trainPoint(float *pointArr, SP_integer pointArrSize, 
				SP_integer label)
{
	// convert the Prolog array to arma::rowvec
	vec pointVector = convertArrayToColvec(pointArr, pointArrSize);
	// check if data Dim fits trained Dim
	if (pointVector.size() != trainedDimensionality)
	{
		raisePrologSystemExeption("The Datapoints Dimensionality doesnt fit the trained Dimensionality!");
		return;
	}

	if(label < 0)
	{
		raisePrologSystemExeption("The given Label should be positive!");
        return;
	}

	
	try
	{
		naiveBayesClassifier.Train(pointVector, label);
	}
	catch(const std::exception& e)
	{
		raisePrologSystemExeption(e.what());
		return;
	}
	isModelTrained = true;
}


// input: 
// output: 	mat	sample variances
// description: 
//			Get the sample variances for each class.
//
void variances(float **variancesMatArr, SP_integer *variancesMatColNum, SP_integer *variancesMatRowNum)
{
	if (!isModelTrained)
	{
		raisePrologSystemExeption("The Model is not trained!");
		return;
	}
	
	// get the ReturnMat
	mat variancesReturnMat = naiveBayesClassifier.Variances();


	// return the Matrix
	returnMatrixInformation(variancesReturnMat, variancesMatArr, variancesMatColNum, variancesMatRowNum);
}
