
:- module(naive_bayes_classifier, [     nbc_initModelWithTrain/6,
                                        nbc_initModelNoTrain/3,
                                          
                                        nbc_classifyMatrix/5,
                                        nbc_means/2,
                                        nbc_probabilities/2,
                                        nbc_trainMatrix/5,
                                        nbc_trainPoint/2,
                                        nbc_variances/2]).

%% requirements of library(struct)
:- load_files(library(str_decl),
              [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses              => 0,
%%              bool    incrementalVariance     => (1)true / (0)false => false,
%%              float32 epsilon                 => 1e-10
%%
%% --Output--
%%
%% --Description--
%%              Initializes the classifier as per the input and then trains it by calculating the sample mean and variances.
%%
nbc_initModelWithTrain(DataList, DataRows, LabelsList, NumClasses, IncrementalVar, Epsilon) :-
        NumClasses >= 0,
        Epsilon >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        initModelWithTrainI(X, Xsize, Xrownum, Y, Ysize, NumClasses, IncrementalVar, Epsilon).

foreign(initModelWithTrain, c, initModelWithTrainI(     +pointer(float_array), +integer, +integer, 
                                                        +pointer(float_array), +integer, 
                                                        +integer, 
                                                        +integer, 
                                                        +float32)).


%% --Input--
%%              int     dimensionality          => 1,
%%              int     numClasses              => 0,
%%              float32 epsilon                 => 1e-10
%%
%% --Output--
%%
%% --Description--
%%              Initialize the Naive Bayes classifier without performing training.
%%              All of the parameters of the model will be initialized to zero. Be sure to use train before calling classify, otherwise the results may be meaningless.
%%
nbc_initModelNoTrain(Dimensionality, NumClasses, Epsilon) :-
        Dimensionality > 0,
        NumClasses >= 0,
        Epsilon >= 0,
        initModelNoTrainI(Dimensionality, NumClasses, Epsilon).

foreign(initModelNoTrain, c, initModelNoTrainI( +integer, 
                                                +integer, 
                                                +float32)).


%% --Input--
%%              mat     data
%%
%% --Output--
%%              vec     predictions,
%%              mat     probabilities
%%
%% --Description--
%%              Classify the given points using the trained NaiveBayesClassifier model and also return estimates of the probabilities for each class in the given matrix.
%%              The predicted labels for each point are stored in the given vector.
%%
%%foreign(classifyPoint, c, classifyPoint(+pointer(float_array), +integer, -integer, -pointer(float_array), -integer)).


%% --Input--
%%              mat     data
%%
%% --Output--
%%              vec     predictions,
%%              mat     probabilities
%%
%% --Description--
%%              Classify the given points using the trained NaiveBayesClassifier model and also return estimates of the probabilities for each class in the given matrix.
%%              The predicted labels for each point are stored in the given vector.
%%
nbc_classifyMatrix(DataList, DataRows, PredictionList, ProbsList, ZCols) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrows, X)),
        classifyMatrixI(X, Xsize, Xrows, Y, Ysize, Z, ZCols, ZRows),
        convert_float_array_to_list(Y, Ysize, PredictionList),
        convert_float_array_to_2d_list(Z, ZCols, ZRows, ProbsList).

foreign(classifyMatrix, c, classifyMatrixI(     +pointer(float_array), +integer, +integer, 
                                                -pointer(float_array), -integer, 
                                                -pointer(float_array), -integer, -integer)).


%% --Input--
%%
%% --Output--
%%              mat     means
%%
%% --Description--
%%              Get the sample means for each class.
%%
nbc_means(MeansList, XCols) :-
        meansI(X, XCols, XRows),
        convert_float_array_to_2d_list(X, XCols, XRows, MeansList).

foreign(means, c, meansI(-pointer(float_array), -integer, -integer)).


%% --Input--
%%
%% --Output--
%%              mat     probabilities
%%
%% --Description--
%%              Get the sample probabilities for each class.
%%
nbc_probabilities(ProbabilitiesList, XCols) :-
        probabilitiesI(X, XCols, XRows),
        convert_float_array_to_2d_list(X, XCols, XRows, ProbabilitiesList).

foreign(probabilities, c, probabilitiesI(-pointer(float_array), -integer, -integer)).


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses              => 0,
%%              bool    incrementalVariance     => (1)true / (0)false => false
%%
%% --Output--
%%
%% --Description--
%%              Train the Naive Bayes classifier on the given dataset. The data must be the same dimensionality as the existing model parameters.
%%
nbc_trainMatrix(DataList, DataRows, LabelsList, NumClasses, IncrementalVar) :-
        NumClasses >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        trainMatrixI(X, Xsize, Xrownum, Y, Ysize, NumClasses, IncrementalVar).

foreign(trainMatrix, c, trainMatrixI(   +pointer(float_array), +integer, +integer, 
                                        +pointer(float_array), +integer, 
                                        +integer, 
                                        +integer)).


%% --Input--
%%              vec     point,
%%              int     label
%%
%% --Output--
%%
%% --Description--
%%              Train the Naive Bayes classifier on the given point.
%%              This will use the incremental algorithm for updating the model parameters. The data must be the same dimensionality as the existing model parameters.
%%
nbc_trainPoint(DataList, Label) :-
        convert_list_to_float_array(DataList, array(Xsize, X)),
        trainPointI(X, Xsize, Label).
        
foreign(trainPoint, c, trainPointI(     +pointer(float_array), +integer, 
                                        +integer)).


%% --Input--
%%
%% --Output--
%%              mat     variances
%%
%% --Description--
%%              Get the sample variances for each class.
%%
nbc_variances(VariancesList, XCols) :-
        variancesI(X, XCols, XRows),
        convert_float_array_to_2d_list(X, XCols, XRows, VariancesList).

foreign(variances, c, variancesI(-pointer(float_array), -integer, -integer)).


%% Defines the functions that get connected from main.cpp
foreign_resource(naive_bayes_classifier, [      initModelWithTrain,
                                                initModelNoTrain,
                                                 
                                                classifyMatrix,
                                                means,
                                                probabilities,
                                                trainMatrix,
                                                trainPoint,
                                                variances]).

:- load_foreign_resource(naive_bayes_classifier).
