#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "pca_glue.h"
#include <mlpack/methods/pca/pca.hpp>
#include <mlpack/core.hpp>

#include <mlpack/methods/pca/decomposition_policies/randomized_svd_method.hpp>
#include <mlpack/methods/pca/decomposition_policies/randomized_block_krylov_method.hpp>
#include <mlpack/methods/pca/decomposition_policies/quic_svd_method.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::pca;


// input:   const bool 	                    scaleData = false,
//          const DecompositionPolicy & 	decomposition = DecompositionPolicy(),
//          const arma::mat & 	            data,
//          arma::mat & 	                transformedData,
//          arma::vec & 	                eigVal,
//          arma::mat & 	                eigvec
// output: 
// description: 
//          Apply Principal Component Analysis to the provided data set.
//
void pca(SP_integer scaleData, char const *decompositionPolicy,
            float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum,
            float **transformedMatArr, SP_integer *transformedMatColNum, SP_integer *transformedMatRowNum,
            float **eigValArr, SP_integer *eigValArrSize,
            float **eigVecMatArr, SP_integer *eigVecMatColNum, SP_integer *eigVecMatRowNum)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);

    // create the ReturnMat
    mat transformedReturnMat;
    // create the ReturnVector
    vec eigValReturnVector;
    // create the ReturnMat
    mat eigVecReturnMat;


    try
    {
        // decide for the decomposition Policy
        if (strcmp(decompositionPolicy, "exact") == 0)
        {
            PCA<ExactSVDPolicy>((scaleData == 1)).Apply(data, transformedReturnMat, eigValReturnVector, eigVecReturnMat);
        }
        else if (strcmp(decompositionPolicy, "randomized") == 0)
        {
            PCA<RandomizedSVDPolicy>((scaleData == 1)).Apply(data, transformedReturnMat, eigValReturnVector, eigVecReturnMat);
        }
        else if (strcmp(decompositionPolicy, "randomized_block_krylov") == 0)
        {
            PCA<RandomizedBlockKrylovSVDPolicy>((scaleData == 1)).Apply(data, transformedReturnMat, eigValReturnVector, eigVecReturnMat);
        }
        else if (strcmp(decompositionPolicy, "quic") == 0)
        {
            PCA<QUICSVDPolicy>((scaleData == 1)).Apply(data, transformedReturnMat, eigValReturnVector, eigVecReturnMat);
        }
        else
        {
            raisePrologDomainExeption(decompositionPolicy, 2, "The given DecompositionPolicy is unkown!", "pca");
            return;
        }
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }


    // return the Matrix
    returnMatrixInformation(transformedReturnMat, transformedMatArr, transformedMatColNum, transformedMatRowNum);
    // return the Vector
    returnVectorInformation(eigValReturnVector, eigValArr, eigValArrSize);
    // return the Matrix
    returnMatrixInformation(eigVecReturnMat, eigVecMatArr, eigVecMatColNum, eigVecMatRowNum);
}


// input:   const bool 	                    scaleData = false,
//          const DecompositionPolicy & 	decomposition = DecompositionPolicy(),
//          arma::mat & 	                data,
//          const size_t 	                newDimension 
// output: 
// description: 
//          Use PCA for dimensionality reduction on the given dataset.
//          Define the new dimensionality of the data with newDimension.
//
double pcaDimReduction(SP_integer scaleData, char const *decompositionPolicy,
                        float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum,
                        SP_integer newDimension,
                        float **transformedMatArr, SP_integer *transformedMatColNum, SP_integer *transformedMatRowNum)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    if(data.n_rows < newDimension)
    {
        raisePrologSystemExeption("NewDimension cant be greater than the Dimensionality of the Data!");
        return 0.0;
    }
    double returnValue;


    try
    {
        // decide for the decomposition Policy
        if (strcmp(decompositionPolicy, "exact") == 0)
        {
            returnValue = PCA<ExactSVDPolicy>((scaleData == 1)).Apply(data, (size_t)newDimension);
        }
        else if (strcmp(decompositionPolicy, "randomized") == 0)
        {
            returnValue = PCA<RandomizedSVDPolicy>((scaleData == 1)).Apply(data, (size_t)newDimension);
        }
        else if (strcmp(decompositionPolicy, "randomized_block_krylov") == 0)
        {
            returnValue = PCA<RandomizedBlockKrylovSVDPolicy>((scaleData == 1)).Apply(data, (size_t)newDimension);
        }
        else if (strcmp(decompositionPolicy, "quic") == 0)
        {
            returnValue = PCA<QUICSVDPolicy>((scaleData == 1)).Apply(data, (size_t)newDimension);
        }
        else
        {
            raisePrologDomainExeption(decompositionPolicy, 2, "The given DecompositionPolicy is unkown!", "pca");
            return 0.0;
        }
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return 0.0;
    }


    // return the Matrix
    returnMatrixInformation(data, transformedMatArr, transformedMatColNum, transformedMatRowNum);

    return returnValue;
}


// input:   const bool 	                    scaleData = false,
//          const DecompositionPolicy & 	decomposition = DecompositionPolicy(),
//          arma::mat & 	                data,
//          const double 	                varRetained 
// output: 
// description: 
//          Use PCA for dimensionality reduction on the given dataset.
//          Define to which variance the data should be reduced to.
//
double pcaVarianceDimReduction(SP_integer scaleData, char const *decompositionPolicy,
                                float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum,
                                double varRetained,
                                float **transformedMatArr, SP_integer *transformedMatColNum, SP_integer *transformedMatRowNum)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);

    double returnValue;


    try
    {
        // decide for the decomposition Policy
        if (strcmp(decompositionPolicy, "exact") == 0)
        {
            returnValue = PCA<ExactSVDPolicy>((scaleData == 1)).Apply(data, varRetained);
        }
        else if (strcmp(decompositionPolicy, "randomized") == 0)
        {
            returnValue = PCA<RandomizedSVDPolicy>((scaleData == 1)).Apply(data, varRetained);
        }
        else if (strcmp(decompositionPolicy, "randomized_block_krylov") == 0)
        {
            returnValue = PCA<RandomizedBlockKrylovSVDPolicy>((scaleData == 1)).Apply(data, varRetained);
        }
        else if (strcmp(decompositionPolicy, "quic") == 0)
        {
            returnValue = PCA<QUICSVDPolicy>((scaleData == 1)).Apply(data, varRetained);
        }
        else
        {
            raisePrologDomainExeption(decompositionPolicy, 2, "The given DecompositionPolicy is unkown!", "pca");
            return 0.0;
        }
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return 0.0;
    }


    // return the Matrix
    returnMatrixInformation(data, transformedMatArr, transformedMatColNum, transformedMatRowNum);

    return returnValue;
}
