
:- module(pca, [pca/9,
                pcaDimReduction/8,
                pcaVarianceDimReduction/8]).

%% requirements of library(struct)
:- load_files(library(str_decl),
        [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function


%% --Input--
%%              bool    scaleData                    => (1)true / (0)false,
%%              string  decompositionPolicy     "exact", "randomized", "randomized_block_krylov", "quic",
%%              mat     data,
%%
%% --Output--
%%              mat     transformedData,
%%              vec     eigValues,
%%              mat     eigVectors
%%
%% --Description--
%%              Apply Principal Component Analysis to the provided data set.
%%
pca(ScaleData, DecompositionPolicy, DataList, DataRows, TransformedList, TDataCols, EigValList, EigVecList, ZCols) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrows, X)),
        pcaI(ScaleData, DecompositionPolicy, X, Xsize, Xrows, TData, TDataCols, TDataRows, Y, Ysize, Z, ZCols, ZRows),
        convert_float_array_to_2d_list(TData, TDataCols, TDataRows, TransformedList),
        convert_float_array_to_list(Y, Ysize, EigValList),
        convert_float_array_to_2d_list(Z, ZCols, ZRows, EigVecList).

foreign(pca, c, pcaI(   +integer, 
                        +string,
                        +pointer(float_array), +integer, +integer,
                        -pointer(float_array), -integer, -integer,
                        -pointer(float_array), -integer,
                        -pointer(float_array), -integer, -integer)).


%% --Input--
%%              bool    scaleData               => (1)true / (0)false,
%%              string  decompositionPolicy     "exact", "randomized", "randomized_block_krylov", "quic",
%%              mat     data,
%%              int     newDimension
%%
%% --Output--
%%              mat     transformedData,
%%              float32 Amount of Variance retained => [0,1]
%%
%% --Description--
%%              Use PCA for dimensionality reduction on the given dataset.
%%              Define the new dimensionality of the data with newDimension.
%%
pcaDimReduction(ScaleData, DecompositionPolicy, DataList, DataRows, NewDim, TransformedList, TDataCols, Variance) :-
        NewDim > 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrows, X)),
        pcaDimReductionI(ScaleData, DecompositionPolicy, X, Xsize, Xrows, NewDim, TData, TDataCols, TDataRows, Variance),
        convert_float_array_to_2d_list(TData, TDataCols, TDataRows, TransformedList).

foreign(pcaDimReduction, c, pcaDimReductionI(   +integer, 
                                                +string,
                                                +pointer(float_array), +integer, +integer,
                                                +integer,
                                                -pointer(float_array), -integer, -integer,
                                                [-float32])).


%% --Input--
%%              bool    scaleData               => (1)true / (0)false,
%%              string  decompositionPolicy     "exact", "randomized", "randomized_block_krylov", "quic",
%%              mat     data,
%%              float32 varRetained             => [0,1]
%%
%% --Output--
%%              mat     transformedData,
%%              float32 Amount of Variance retained => [0,1]
%%
%% --Description--
%%              Use PCA for dimensionality reduction on the given dataset.
%%              Define to which variance the data should be reduced to.
%%
pcaVarianceDimReduction(ScaleData, DecompositionPolicy, DataList, DataRows, VarRetained, TransformedList, TDataCols, Variance) :-
        VarRetained >= 0.0,
        VarRetained =< 1.0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrows, X)),
        pcaVarianceDimReductionI(ScaleData, DecompositionPolicy, X, Xsize, Xrows, VarRetained, TData, TDataCols, TDataRows, Variance),
        convert_float_array_to_2d_list(TData, TDataCols, TDataRows, TransformedList).

foreign(pcaVarianceDimReduction, c, pcaVarianceDimReductionI(   +integer, 
                                                                +string,
                                                                +pointer(float_array), +integer, +integer,
                                                                +float32,
                                                                -pointer(float_array), -integer, -integer,
                                                                [-float32])).


%% Defines the functions that get connected from main.cpp
foreign_resource(pca, [ pca,
                        pcaDimReduction,
                        pcaVarianceDimReduction]).

:- load_foreign_resource(pca).
