#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "perceptron_glue.h"
#include <mlpack/methods/perceptron/perceptron.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::perceptron;

// Global Variable of the GlobalMethodObject object so it can be accessed from all functions
Perceptron perceptronGlobal;


// input:   const size_t 	numClasses = 0,
//          const size_t 	dimensionality = 0,
//          const size_t 	maxIterations = 1000
// output: 
// description: 
//          Initilizes the perceptron model and its weight matrix but doesnt train it.
//
void initModelNoTrain(SP_integer numClasses, SP_integer dimensionality, SP_integer maxIterations)
{
   try
   {
        perceptronGlobal = Perceptron(numClasses, dimensionality, maxIterations);
   }
   catch(const std::exception& e)
   {
        raisePrologSystemExeption(e.what());
        return;
   }
}


// input:   const MatType & 	            data,
//          const arma::Row< size_t > &     labels,
//          const size_t 	                numClasses,
//          const size_t 	                maxIterations = 1000
// output: 
// description: 
//          Initilizes the perceptron model and its weight matrix and trains it with the given data.
//
void initModelWithTrain(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
                        float *labelsArr, SP_integer labelsArrSize, 
                        SP_integer numClasses, SP_integer maxIterations)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    // check if labels and weights fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }
    
    // convert the Prolog array to arma::rowvec
    Row< size_t > labelsVector = convertArrayToVec(labelsArr, labelsArrSize);


    try
    {
        perceptronGlobal = Perceptron(data, labelsVector, numClasses, maxIterations);
    }
    catch(const std::out_of_range& e)
    {
        raisePrologSystemExeption("The given Labels dont fit the format [0,Numclasses-1]!");
        return;
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
}


// input: 
// output:  arma::vec&  biases
// description: 
//          Get the biases.
//
void biases(float **biasesArr, SP_integer *biasesArrSize)
{
    // create the ReturnVector
    vec biasesReturnVector = perceptronGlobal.Biases();
    
    // return Vector
    returnVectorInformation(biasesReturnVector, biasesArr, biasesArrSize);
}


// input:   const MatType &         test,
//          arma::Row< size_t > &   predictedLabels <-
// output: 
// description: 
//          After training, use the weights matrix to classify test, and put the predicted classes in predictedLabels.
//
void classify(float *testMatArr, SP_integer testMatSize, SP_integer testMatRowNum, 
                float **predictLabelsArr, SP_integer *predictLabelsArrSize)
{
    // convert the Prolog arrays to arma::mat
    mat test = convertArrayToMat(testMatArr, testMatSize, testMatRowNum);
    
    // create the ReturnVector
    // need to be initialized with the amount of test cols or it will cause in classify a index out of bounds error
    Row< size_t > predictLabelsReturnVector(test.n_cols);


    try
    {
        perceptronGlobal.Classify(test, predictLabelsReturnVector);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    

    // return the Vector
    returnVectorInformation(predictLabelsReturnVector, predictLabelsArr, predictLabelsArrSize);
}


// input:   const MatType & 	            data,
//          const arma::Row< size_t > &     labels,
//          const size_t 	                numClasses,
//          const arma::rowvec & 	        instanceWeights = arma::rowvec()
// output: 
// description: 
//          Train the perceptron on the given data for up to the maximum number of iterations. This training does not reset the model weights, so you can call train/8 on multiple datasets sequentially.
//
void train(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
                float *labelsArr, SP_integer labelsArrSize, 
                SP_integer numClasses, 
                float *instanceWeightsArr, SP_integer instanceWeightsArrSize)
{
    // convert the Prolog arrays to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    // check if labels and weights fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }
    if (data.n_cols != instanceWeightsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of weights!");
        return;
    }
    
    // convert the Prolog array
    Row< size_t > labelsVector = convertArrayToVec(labelsArr, labelsArrSize);
    rowvec instanceWeightsVector = convertArrayToRowvec(instanceWeightsArr, instanceWeightsArrSize);


    // use the model function
    try
    {
        perceptronGlobal.Train(data, labelsVector, numClasses, instanceWeightsVector);
    }
    catch(const std::out_of_range& e)
    {
        raisePrologSystemExeption("The given Labels dont fit the format [0,Numclasses-1]!");
        return;
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
}


// input: 
// output:  const arma::mat&    weights
// description: 
//          Get the weight matrix.
//
void weights(float **weightsMatArr, SP_integer *weightsMatColNum, SP_integer *weightsMatRowNum)
{
    // create the ReturnMat
    mat weightsReturnMat = perceptronGlobal.Weights();
    
    
    // return the Mat
    returnMatrixInformation(weightsReturnMat, weightsMatArr, weightsMatColNum, weightsMatRowNum);
}
