
:- module(perceptron, [ perceptron_initModelNoTrain/3,
                        perceptron_initModelWithTrain/5,
                        perceptron_biases/1,
                        perceptron_classify/3,
                        perceptron_train/5,
                        perceptron_weights/2]).

%% requirements of library(struct)
:- load_files(library(str_decl),
                [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function


%% --Input--
%%              int     numClasses      => 0,
%%              int     dimensionality  => 0,
%%              int     maxIterations   => 1000
%%
%% --Output--: 
%% --Description-- 
%%              Initilizes the perceptron model and its weight matrix but doesnt train it.
%%
perceptron_initModelNoTrain(NumClasses, Dimensionality, MaxIterations) :-
        NumClasses >= 0,
        Dimensionality > 0,
        MaxIterations >= 0,
        initModelNoTrainI(NumClasses, Dimensionality, MaxIterations).

foreign(initModelNoTrain, c, initModelNoTrainI(+integer, +integer, +integer)).


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses,
%%              int     maxIterations   => 1000
%%
%% --Output--
%% --Description-- 
%%              Initilizes the perceptron model and its weight matrix and trains it with the given data.
%%
perceptron_initModelWithTrain(DataList, DataRows, LabelsList, NumClasses, MaxIterations) :-
        NumClasses >= 0,
        MaxIterations >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        initModelWithTrainI(X, Xsize, Xrownum, Y, Ysize, NumClasses, MaxIterations).

foreign(initModelWithTrain, c, initModelWithTrainI(     +pointer(float_array), +integer, +integer, 
                                                        +pointer(float_array), +integer, 
                                                        +integer, +integer)).


%% --Input--    
%%
%% --Output--
%%              vec     biases          => biases(float_array), biasesSize
%%
%% --Description-- 
%%              Get the biases.
%%
perceptron_biases(BiasesList) :-
        biasesI(Y, Ysize),
        convert_float_array_to_list(Y, Ysize, BiasesList).

foreign(biases, c, biasesI(-pointer(float_array), -integer)).

%% --Input--
%%              mat     test
%%
%% --Output--
%%              vec     predicLabels
%%
%% --Description-- 
%%              After training, use the weights matrix to classify test, and put the predicted classes in predictedLabels.
%%
perceptron_classify(TestList, TestRows, PredictLabelList) :-
        convert_list_to_float_array(TestList, TestRows, array(Xsize, Xrows, X)),
        classifyI(X, Xsize, Xrows, Y, Ysize),
        convert_float_array_to_list(Y, Ysize, PredictLabelList).

foreign(classify, c, classifyI( +pointer(float_array), +integer, +integer, 
                                -pointer(float_array), -integer)).


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses,
%%              vec     instWeights
%%
%% --Output--
%% --Description--
%%              Train the perceptron on the given data for up to the maximum number of iterations. This training does not reset the model weights, so you can call train/8 on multiple datasets sequentially.
%%
perceptron_train(DataList, DataRows, LabelsList, NumClasses, WeightsList) :-
        NumClasses >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        convert_list_to_float_array(WeightsList, array(Zsize, Z)),
        trainI(X, Xsize, Xrownum, Y, Ysize, NumClasses, Z, Zsize).

foreign(train, c, trainI(       +pointer(float_array), +integer, +integer, 
                        	+pointer(float_array), +integer, 
                        	+integer, 
                        	+pointer(float_array), +integer)).


%% --Input--
%%
%% --Output--
%%              mat     weights
%%
%% --Description--
%%              Get the weight matrix.
%%
perceptron_weights(WeightsList, XCols) :-
        weightsI(X, XCols, XRows),
        convert_float_array_to_2d_list(X, XCols, XRows, WeightsList).

foreign(weights, c, weightsI(-pointer(float_array), -integer, -integer)).


%% Defines the functions that get connected from main.cpp
foreign_resource(perceptron, [  initModelNoTrain,
                                initModelWithTrain,
                                biases,
                                classify,
                                train,
                                weights]).

:- load_foreign_resource(perceptron).
