#include <sicstus/sicstus.h>
/* ex_glue.h is generated by splfr from the foreign/[2,3] facts.
   Always include the glue header in your foreign resource code.
*/
#include "random_forest_glue.h"
#include <mlpack/methods/random_forest/random_forest.hpp>
#include <mlpack/core.hpp>

// including helper functions for converting between arma structures and arrays
#include "../../helper_files/helper.hpp"

// some of the most used namespaces
using namespace arma;
using namespace mlpack;
using namespace std;
using namespace mlpack::tree;

// Global Variable of the RandomForest object so it can be accessed from all functions
RandomForest randomForest;


// input: 
// output: 
// description: 
//          Initilizes the model without training it.
//
void initModelNoTrain(SP_integer)
{
    randomForest = RandomForest();
}


// input:   const MatType & 	            dataset,
//          const arma::Row< size_t > & 	labels,
//          const size_t 	                numClasses,
//          const size_t 	                numTrees = 20,
//          const size_t 	                minimumLeafSize = 1,
//          const double 	                minimumGainSplit = 1e-7,
//          const size_t 	                maximumDepth = 0,
//          DimensionSelectionType 	        dimensionSelector = DimensionSelectionType()
// output: 
// description: 
//          Initilizes the model and trains it but does not apply weights to it.
//
void initModelWithTrainNoWeights(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
                                    float *labelsArr, SP_integer labelsArrSize, 
                                    SP_integer numClasses, SP_integer numTrees, SP_integer minimumLeafSize, double minimumGainSplit, SP_integer maximumDepth)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    // check if labels fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }
    // convert the Prolog array to arma::rowvec
    Row< size_t > labelsVector = convertArrayToVec(labelsArr, labelsArrSize);
    

    try
    {
        randomForest = RandomForest(data, labelsVector, numClasses, numTrees, minimumLeafSize, minimumGainSplit, maximumDepth);
    }
    catch(const std::out_of_range& e)
    {
        raisePrologSystemExeption("The given Labels dont fit the format [0,Numclasses-1]!");
        return;
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
}


// input:   onst MatType & 	                dataset,
//          const arma::Row< size_t > & 	labels,
//          const size_t 	                numClasses,
//          const arma::rowvec & 	        weights,
//          const size_t 	                numTrees = 20,
//          const size_t 	                minimumLeafSize = 1,
//          const double 	                minimumGainSplit = 1e-7,
//          const size_t 	                maximumDepth = 0,
//          DimensionSelectionType          dimensionSelector = DimensionSelectionType()
// output: 
// description: 
//          Initilizes the model, trains it and applies weights to it.
//
void initModelWithTrainWithWeights(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
                                    float *labelsArr, SP_integer labelsArrSize, 
                                    SP_integer numClasses,
									float *weightsArr, SP_integer weightsArrSize, 
                                    SP_integer numTrees, SP_integer minimumLeafSize, double minimumGainSplit, SP_integer maximumDepth)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    // check if labels and weights fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return;
    }
    if (data.n_cols != weightsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of weights!");
        return;
    }
    
    // convert the Prolog array to arma::Row< size_t >
    Row< size_t > labelsVector = convertArrayToVec(labelsArr, labelsArrSize);
    // convert the Prolog array to arma::rowvec
    rowvec weightsVector = convertArrayToRowvec(weightsArr, weightsArrSize);
    
    
    try
    {
        randomForest = RandomForest(data, labelsVector, numClasses, weightsVector, numTrees, minimumLeafSize, minimumGainSplit, maximumDepth);
    }
    catch(const std::out_of_range& e)
    {
        raisePrologSystemExeption("The given Labels dont fit the format [0,Numclasses-1]!");
        return;
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
}


// input:   const VecType &     point,
//          size_t & 	        prediction <-,
//          arma::vec & 	    probabilities <-
// output: 
// description: 
//          Predict the class of the given point and return the predicted class probabilities for each class.
//
void classifyPoint(float *pointArr, SP_integer pointArrSize, 
                    SP_integer *prediction, 
                    float **probsArr, SP_integer *probsArrSize)
{
    // convert the Prolog array to arma::rowvec
    rowvec pointVector = convertArrayToRowvec(pointArr, pointArrSize);

    // create the ReturnVector
    vec probsReturnVector;

    size_t predicReturn;


    try
    {
        randomForest.Classify(pointVector, predicReturn, probsReturnVector);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    

    // return the predic value
    *prediction = predicReturn;

    // return the Vector
    returnVectorInformation(probsReturnVector, probsArr, probsArrSize);
}


// input:   const MatType & 	        data,
//          arma::Row< size_t > & 	    predictions <-,
//          arma::mat & 	            probabilities <-
// output: 
// description: 
//          Predict the classes of each point in the given dataset, also returning the predicted class probabilities for each point.
//
void classifyMatrix(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
                    float **predicArr, SP_integer *predicArrSize, 
                    float **probsMatArr, SP_integer *probsMatColNum, SP_integer *probsMatRowNum)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);

    // create the ReturnVector
    Row< size_t > predicReturnVector;
    // create the ReturnMat
    mat probsReturnMat;


    try
    {
        randomForest.Classify(data, predicReturnVector, probsReturnMat);
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return;
    }
    

    // return the Vector
    returnVectorInformation(predicReturnVector, predicArr, predicArrSize);
    // return the Matrix
    returnMatrixInformation(probsReturnMat, probsMatArr, probsMatColNum, probsMatRowNum);
}


// input: 
// output:  const size_t   number of trees
// description: 
//          Get the number of trees in the forest.
//
SP_integer numTrees()
{
    return randomForest.NumTrees();
}


// TODO:    doesnt accept warmStart parameter
//
// input:   const MatType & 	            data,
//          const arma::Row< size_t > & 	labels,
//          const size_t 	                numClasses,
//          const size_t 	                numTrees = 20,
//          const size_t 	                minimumLeafSize = 1,
//          const double 	                minimumGainSplit = 1e-7,
//          const size_t 	                maximumDepth = 0,
//          const bool                      warmStart = false,
//          DimensionSelectionType 	        dimensionSelector = DimensionSelectionType()
//
// output:  double                          average entropy of all trees trained
// description: 
//          Train the random forest on the given labeled training data with the given number of trees.
//
double trainNoWeights(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, 
                        float *labelsArr, SP_integer labelsArrSize, 
                        SP_integer numClasses, SP_integer numTrees, SP_integer minimumLeafSize, double minimumGainSplit, SP_integer maximumDepth)
{
    // convert the Prolog array to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    // check if labels fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }
    // convert the Prolog array to arma::rowvec
    Row< size_t > labelsVector = convertArrayToVec(labelsArr, labelsArrSize);

    
    try
    {
        return randomForest.Train(data, labelsVector, numClasses, numTrees, minimumLeafSize, minimumGainSplit, maximumDepth);
    }
    catch(const std::out_of_range& e)
    {
        raisePrologSystemExeption("The given Labels dont fit the format [0,Numclasses-1]!");
        return 0.0;
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return 0.0;
    }
}


// TODO:    doesnt accept warmStart parameter
//
// input:   const MatType & 	            data,
//          const arma::Row< size_t > & 	labels,
//          const size_t 	                numClasses,
//          const arma::rowvec & 	        weights,
//          const size_t 	                numTrees = 20,
//          const size_t 	                minimumLeafSize = 1,
//          const double 	                minimumGainSplit = 1e-7,
//          const size_t 	                maximumDepth = 0,
//          const bool 	                    warmStart = false,
//          DimensionSelectionType 	        dimensionSelector = DimensionSelectionType()
//
// output:  double                          average entropy of all trees trained
// description: 
//          Train the random forest on the given weighted labeled training data with the given number of trees.
//
double trainWithWeights(float *dataMatArr, SP_integer dataMatSize, SP_integer dataMatRowNum, float *labelsArr, SP_integer labelsArrSize, SP_integer numClasses,
							float *weightsArr, SP_integer weightsArrSize, SP_integer numTrees, SP_integer minimumLeafSize, double minimumGainSplit, SP_integer maximumDepth)
{
    // convert the Prolog arrays to arma::mat
    mat data = convertArrayToMat(dataMatArr, dataMatSize, dataMatRowNum);
    // check if labels and weights fit the data
    if (data.n_cols != labelsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of labels!");
        return 0.0;
    }
    if (data.n_cols != weightsArrSize)
    {
        raisePrologSystemExeption("The number of data points does not match the number of weights!");
        return 0.0;
    }
    // convert the Prolog arrays to arma::rowvec
    Row< size_t > labelsVector = convertArrayToVec(labelsArr, labelsArrSize);
    // convert the Prolog arrays to arma::rowvec
    rowvec weightsVector = convertArrayToRowvec(weightsArr, weightsArrSize);
    
    
    try
    {
        return randomForest.Train(data, labelsVector, numClasses, weightsVector, numTrees, minimumLeafSize, minimumGainSplit, maximumDepth);
    }
    catch(const std::out_of_range& e)
    {
        raisePrologSystemExeption("The given Labels dont fit the format [0,Numclasses-1]!");
        return 0.0;
    }
    catch(const std::exception& e)
    {
        raisePrologSystemExeption(e.what());
        return 0.0;
    }
}
