
:- module(random_forest, [      random_forest_initModelNoTrain/0,
                                random_forest_initModelWithTrainNoWeights/8,
                                random_forest_initModelWithTrainWithWeights/9,
                                random_forest_classifyPoint/3,
                                random_forest_classifyMatrix/5,
                                random_forest_numTrees/1,
                                random_forest_trainNoWeights/9,
                                random_forest_trainWithWeights/10]).

%% requirements of library(struct)
:- load_files(library(str_decl),
                [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected function


%% --Input--
%%              int     notUsedParam
%%
%% --Output--
%%
%% --Description--
%%              Initilizes the model without training it.
%%
random_forest_initModelNoTrain :-
        initModelNoTrainI(0).

foreign(initModelNoTrain, c, initModelNoTrainI(+integer)).


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses              => 0,
%%              int     numTrees                => 20,
%%              int     minimumLeafSize         => 1,
%%              float32 minimumGainSplit        => 1e-7,
%%              int     maximumDepth            => 0
%%
%% --Output--
%%
%% --Description--
%%              Initilizes the model and trains it but does not apply weights to it.
%%
random_forest_initModelWithTrainNoWeights(DataList, DataRows, LabelsList, NumClasses, NumTrees, MinimumLeafSize, MinimumGainSplit, MaximumDepth) :-
        NumClasses >= 0,
        NumTrees > 0,
        MinimumLeafSize > 0,
        MinimumGainSplit >= 0.0,
        MaximumDepth >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        initModelWithTrainNoWeightsI(X, Xsize, Xrownum, Y, Ysize, NumClasses, NumTrees, MinimumLeafSize, MinimumGainSplit, MaximumDepth).

foreign(initModelWithTrainNoWeights, c, initModelWithTrainNoWeightsI(    
        +pointer(float_array), +integer, +integer, 
        +pointer(float_array), +integer, 
        +integer, +integer, +integer, +float32, +integer)).


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses              => 0,
%%              vec     weights,
%%              int     numTrees                => 20,
%%              int     minimumLeafSize         => 1,
%%              float32 minimumGainSplit        => 1e-7,
%%              int     maximumDepth            => 0
%%
%% --Output--
%%
%% --Description--
%%              Initilizes the model, trains it and applies weights to it.
%%
random_forest_initModelWithTrainWithWeights(DataList, DataRows, LabelsList, NumClasses, WeightsList, NumTrees, MinimumLeafSize, MinimumGainSplit, MaximumDepth) :-
        NumClasses >= 0,
        NumTrees > 0,
        MinimumLeafSize > 0,
        MinimumGainSplit >= 0.0,
        MaximumDepth >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        convert_list_to_float_array(WeightsList, array(Zsize, Z)),
        initModelWithTrainWithWeightsI(X, Xsize, Xrownum, Y, Ysize, NumClasses, Z, Zsize, NumTrees, MinimumLeafSize, MinimumGainSplit, MaximumDepth).

foreign(initModelWithTrainWithWeights, c, initModelWithTrainWithWeightsI(
        +pointer(float_array), +integer, +integer, 
        +pointer(float_array), +integer, 
        +integer, 
        +pointer(float_array), +integer, 
        +integer , +integer, +float32, +integer)).


%% --Input--
%%              vec     point
%%
%% --Output--
%%              int     prediction,
%%              vec     probabilities
%%
%% --Description--
%%              Predict the class of the given point and return the predicted class probabilities for each class.
%%              Random forest has to be train before using this.
%%
random_forest_classifyPoint(DataList, Prediction, AssignList) :-
        convert_list_to_float_array(DataList, array(Xsize, X)),
        classifyPointI(X, Xsize, Prediction, Y, Ysize),
        convert_float_array_to_list(Y, Ysize, AssignList).

foreign(classifyPoint, c, classifyPointI(       +pointer(float_array), +integer, 
                                        	-integer, 
                                        	-pointer(float_array), -integer)).


%% --Input--
%%              mat     data
%%
%% --Output--
%%              vec     prediction,
%%              mat     probabilities
%%
%% --Description--
%%              Predict the classes of each point in the given dataset, also returning the predicted class probabilities for each point.
%%              Random forest has to be train before using this.
%%
random_forest_classifyMatrix(DataList, DataRows, PredictionList, ProbsList, ZCols) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrows, X)),
        classifyMatrixI(X, Xsize, Xrows, Y, Ysize, Z, ZCols, ZRows),
        convert_float_array_to_list(Y, Ysize, PredictionList),
        convert_float_array_to_2d_list(Z, ZCols, ZRows, ProbsList).

foreign(classifyMatrix, c, classifyMatrixI(     +pointer(float_array), +integer, +integer, 
                                                -pointer(float_array), -integer, 
                                                -pointer(float_array), -integer, -integer)).


%% --Input--
%%
%% --Output--
%%              int     numTrees
%%
%% --Description--
%%              Get the number of trees in the forest.
%%
random_forest_numTrees(NumTrees) :-
        numTreesI(NumTrees).
foreign(numTrees, c, numTreesI([-integer])).


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses              => 0,
%%              int     numTrees                => 20,
%%              int     minimumLeafSize         => 1,
%%              float32 minimumGainSplit        => 1e-7,
%%              int     maximumDepth            => 0
%%
%% --Output--
%%              float32 average entropy of all the decision trees
%%
%% --Description--
%%              Train the random forest on the given labeled training data with the given number of trees.
%%              The minimumLeafSize and minimumGainSplit parameters are given to each individual decision tree during tree building.
%%
random_forest_trainNoWeights(DataList, DataRows, LabelsList, NumClasses, NumTrees, MinimumLeafSize, MinimumGainSplit, MaximumDepth, Entropy) :-
        NumClasses >= 0,
        NumTrees > 0,
        MinimumLeafSize > 0,
        MinimumGainSplit >= 0.0,
        MaximumDepth >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        trainNoWeightsI(X, Xsize, Xrownum, Y, Ysize, NumClasses, NumTrees, MinimumLeafSize, MinimumGainSplit, MaximumDepth, Entropy).
        
foreign(trainNoWeights, c, trainNoWeightsI(     +pointer(float_array), +integer, +integer, 
                                                +pointer(float_array), +integer, 
                                                +integer, +integer, +integer, +float32, +integer, 
                                                [-float32])).


%% --Input--
%%              mat     data,
%%              vec     labels,
%%              int     numClasses              => 0,
%%              vec     weights,
%%              int     numTrees                => 20,
%%              int     minimumLeafSize         => 1,
%%              float32 minimumGainSplit        => 1e-7,
%%              int     maximumDepth            => 0
%%
%% --Output--
%%              float32 average entropy of all the decision trees
%%
%% --Description--
%%              Train the random forest on the given weighted labeled training data with the given number of trees.
%%              The minimumLeafSize and minimumGainSplit parameters are given to each individual decision tree during tree building.
%%
random_forest_trainWithWeights(DataList, DataRows, LabelsList, NumClasses, WeightsList, NumTrees, MinimumLeafSize, MinimumGainSplit, MaximumDepth, Entropy) :-
        NumClasses >= 0,
        NumTrees > 0,
        MinimumLeafSize > 0,
        MinimumGainSplit >= 0.0,
        MaximumDepth >= 0,
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        convert_list_to_float_array(LabelsList, array(Ysize, Y)),
        convert_list_to_float_array(WeightsList, array(Zsize, Z)),
        trainWithWeightsI(X, Xsize, Xrownum, Y, Ysize, NumClasses, Z, Zsize, NumTrees, MinimumLeafSize, MinimumGainSplit, MaximumDepth, Entropy).

foreign(trainWithWeights, c, trainWithWeightsI( +pointer(float_array), +integer, +integer, 
                                                +pointer(float_array), +integer, 
                                                +integer, 
                                                +pointer(float_array), +integer, 
                                                +integer, +integer, +float32, +integer, 
                                                [-float32])).


%% Defines the functions that get connected from main.cpp
foreign_resource(random_forest, [       initModelNoTrain,
                                        initModelWithTrainNoWeights,
                                        initModelWithTrainWithWeights,
                                        classifyPoint,
                                        classifyMatrix,
                                        numTrees,
                                        trainNoWeights,
                                        trainWithWeights]).

:- load_foreign_resource(random_forest).
