
:- module(sparse_coding, [      sparse_coding_initModelWithTrain/8,
                                sparse_coding_initModelNoTrain/6,
                                sparse_coding_encode/4,
                                sparse_coding_train/3]).

%% requirements of library(struct)
:- load_files(library(str_decl),
        [when(compile_time), if(changed)]).

%% needed for using the array type
:- use_module(library(structs)).
:- use_module('../../helper_files/helper.pl').

%% type definitions for the float array
:- foreign_type
        float32          = float_32,
        float_array      = array(float32).

%% definitions for the connected functionsparse_coding_




%% --Input--
%%              mat     data,
%%              int     atoms                   => 15,
%%              float32 lambda1                 => 0,
%%              float32 lambda2                 => 0,
%%              int     maxIterations           => 0,
%%              float32 objTolerance            => 0.01,
%%              float32 newtonTolerance         => 1e-6,
%%
%% --Output--
%%
%% --Description--
%%              Initializes sparse_coding model and trains it.
%%
sparse_coding_initModelWithTrain(MatList, MatRows, Atoms, Lambda1, Lambda2, MaxIterations, ObjTolerance, NewtonTolerance) :-
        Atoms >= 0,
        Lambda1 > 0,
        Lambda2 >= 0,
        MaxIterations >= 0,
        ObjTolerance > 0,
        NewtonTolerance > 0,
        convert_list_to_float_array(MatList, MatRows, array(Xsize, Xrownum, X)),
        initModelWithTrainI(X, Xsize, Xrownum, Atoms, Lambda1, Lambda2, MaxIterations, ObjTolerance, NewtonTolerance).

foreign(initModelWithTrain, c, initModelWithTrainI(     +pointer(float_array), +integer, +integer,
                                                        +integer, +float32, +float32, +integer, +float32, +float32)).



%% --Input--
%%              int     atoms                   => 15,
%%              float32 lambda1                 => 0,
%%              float32 lambda2                 => 0,
%%              int     maxIterations           => 0,
%%              float32 objTolerance            => 0.01,
%%              float32 newtonTolerance         => 1e-6,
%%
%% --Output--
%%
%% --Description--
%%              Initializes sparse_coding model but will not train the model, and a subsequent call to Train will be required before the model can encode points with Encode.
%%
sparse_coding_initModelNoTrain(Atoms, Lambda1, Lambda2, MaxIterations, ObjTolerance, NewtonTolerance) :-
        Atoms >= 0,
        Lambda1 > 0,
        Lambda2 >= 0,
        MaxIterations >= 0,
        ObjTolerance > 0,
        NewtonTolerance > 0,
        initModelNoTrainI(Atoms, Lambda1, Lambda2, MaxIterations, ObjTolerance, NewtonTolerance).

foreign(initModelNoTrain, c, initModelNoTrainI(+integer, +float32, +float32, +integer, +float32, +float32)).



%% --Input--
%%              mat     data
%%
%% --Output--
%%              mat     codes
%%
%% --Description--
%%              Sparse code each point in the given dataset via LARS, using the current dictionary and store the encoded data in the codes matrix.
%%
sparse_coding_encode(DataList, DataRows, CodesList, YCols) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        encodeI(X, Xsize, Xrownum, Y, YCols, YRows),
        convert_float_array_to_2d_list(Y, YCols, YRows, CodesList).

foreign(encode, c, encodeI(     +pointer(float_array), +integer, +integer, 
                                -pointer(float_array), -integer, -integer)).



%% --Input--
%%              mat     data
%%
%% --Output--
%%              float32 final objective value
%%
%% --Description--
%%              Train the sparse coding model on the given dataset.
%%
sparse_coding_train(DataList, DataRows, ReturnValue) :-
        convert_list_to_float_array(DataList, DataRows, array(Xsize, Xrownum, X)),
        trainI(X, Xsize, Xrownum, ReturnValue).

foreign(train, c, trainI(       +pointer(float_array), +integer, +integer, 
                                [-float32])).


%% Defines the functions that get connected from main.cpp
foreign_resource(sparse_coding, [       initModelWithTrain,
                                        initModelNoTrain,
                                        encode,
                                        train]).

:- load_foreign_resource(sparse_coding).
